'''Unit tests for QTrapGroup.'''
import unittest
import numpy as np
from pyqtgraph.Qt import QtCore, QtWidgets, QtTest
from QHOT.lib.traps.QTrap import QTrap
from QHOT.lib.traps.QTrapGroup import QTrapGroup

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class TestAddTrap(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.))

    def test_single_sets_parent(self):
        trap = QTrap(phase=0.)
        self.group.addTrap(trap)
        self.assertIs(trap.parent(), self.group)

    def test_single_increments_len(self):
        self.group.addTrap(QTrap(phase=0.))
        self.assertEqual(len(self.group), 1)

    def test_list_sets_parents(self):
        traps = [QTrap(phase=0.) for _ in range(3)]
        self.group.addTrap(traps)
        self.assertTrue(all(t.parent() is self.group for t in traps))

    def test_list_increments_len(self):
        traps = [QTrap(phase=0.) for _ in range(3)]
        self.group.addTrap(traps)
        self.assertEqual(len(self.group), 3)

    def test_iter_after_add(self):
        trap = QTrap(phase=0.)
        self.group.addTrap(trap)
        self.assertEqual(list(self.group), [trap])


class TestRemoveTrap(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.))

    def test_removes_child(self):
        trap = QTrap(phase=0.)
        self.group.addTrap(trap)
        self.group.removeTrap(trap)
        self.assertEqual(len(self.group), 0)

    def test_clears_parent(self):
        trap = QTrap(phase=0.)
        self.group.addTrap(trap)
        self.group.removeTrap(trap)
        self.assertIsNot(trap.parent(), self.group)

    def test_noop_if_not_child(self):
        trap = QTrap(phase=0.)
        other = QTrapGroup(r=(5., 5., 0.))
        other.addTrap(trap)
        self.group.removeTrap(trap)
        self.assertIs(trap.parent(), other)


class TestTrapsAndIter(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.))
        self.traps = [QTrap(phase=0.) for _ in range(3)]
        self.group.addTrap(self.traps)

    def test_traps_property(self):
        self.assertEqual(self.group.traps, self.traps)

    def test_iter_yields_children(self):
        self.assertEqual(list(self.group), self.traps)

    def test_empty_group_iter(self):
        self.assertEqual(list(QTrapGroup(r=(0., 0., 0.))), [])

    def test_len(self):
        self.assertEqual(len(self.group), 3)


class TestRSetter(unittest.TestCase):

    def test_group_position_updates(self):
        group = QTrapGroup(r=(0., 0., 0.))
        group.addTrap(QTrap(r=(1., 0., 0.), phase=0.))
        group.r = (3., 0., 0.)
        np.testing.assert_array_equal(group.r, [3., 0., 0.])

    def test_child_translates_by_delta(self):
        group = QTrapGroup(r=(0., 0., 0.))
        trap = QTrap(r=(1., 0., 0.), phase=0.)
        group.addTrap(trap)
        group.r = (3., 0., 0.)
        np.testing.assert_array_almost_equal(trap.r, [4., 0., 0.])

    def test_geometry_preserved(self):
        group = QTrapGroup(r=(0., 0., 0.))
        t1 = QTrap(r=(1., 0., 0.), phase=0.)
        t2 = QTrap(r=(-1., 0., 0.), phase=0.)
        group.addTrap([t1, t2])
        group.r = (5., 0., 0.)
        np.testing.assert_array_almost_equal(t1.r, [6., 0., 0.])
        np.testing.assert_array_almost_equal(t2.r, [4., 0., 0.])

    def test_emits_changed(self):
        group = QTrapGroup(r=(0., 0., 0.))
        group.addTrap(QTrap(r=(1., 0., 0.), phase=0.))
        spy = QtTest.QSignalSpy(group.changed)
        group.r = (3., 0., 0.)
        self.assertEqual(len(spy), 1)

    def test_nested_group_translates(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(2., 0., 0.))
        trap = QTrap(r=(3., 0., 0.), phase=0.)
        inner.addTrap(trap)
        outer.addTrap(inner)
        outer.r = (1., 0., 0.)
        np.testing.assert_array_almost_equal(outer.r, [1., 0., 0.])
        np.testing.assert_array_almost_equal(inner.r, [3., 0., 0.])
        np.testing.assert_array_almost_equal(trap.r, [4., 0., 0.])


class TestIsWithin(unittest.TestCase):

    def setUp(self):
        self.rect = QtCore.QRectF(0., 0., 10., 10.)

    def test_all_inside(self):
        group = QTrapGroup(r=(5., 5., 0.))
        group.addTrap([QTrap(r=(3., 3., 0.), phase=0.),
                       QTrap(r=(7., 7., 0.), phase=0.)])
        self.assertTrue(group.isWithin(self.rect))

    def test_all_outside(self):
        group = QTrapGroup(r=(15., 5., 0.))
        group.addTrap([QTrap(r=(12., 3., 0.), phase=0.),
                       QTrap(r=(14., 7., 0.), phase=0.)])
        self.assertFalse(group.isWithin(self.rect))

    def test_straddles_boundary(self):
        group = QTrapGroup(r=(5., 5., 0.))
        group.addTrap([QTrap(r=(3., 3., 0.), phase=0.),
                       QTrap(r=(15., 7., 0.), phase=0.)])
        self.assertFalse(group.isWithin(self.rect))

    def test_nested_all_inside(self):
        outer = QTrapGroup(r=(5., 5., 0.))
        inner = QTrapGroup(r=(4., 4., 0.))
        inner.addTrap([QTrap(r=(3., 3., 0.), phase=0.),
                       QTrap(r=(5., 5., 0.), phase=0.)])
        outer.addTrap(inner)
        self.assertTrue(outer.isWithin(self.rect))

    def test_nested_straddles_boundary(self):
        outer = QTrapGroup(r=(5., 5., 0.))
        inner = QTrapGroup(r=(4., 4., 0.))
        inner.addTrap([QTrap(r=(3., 3., 0.), phase=0.),
                       QTrap(r=(15., 5., 0.), phase=0.)])
        outer.addTrap(inner)
        self.assertFalse(outer.isWithin(self.rect))


class TestLeaves(unittest.TestCase):

    def test_flat_group_yields_traps(self):
        group = QTrapGroup(r=(0., 0., 0.))
        traps = [QTrap(phase=0.) for _ in range(3)]
        group.addTrap(traps)
        self.assertEqual(list(group.leaves()), traps)

    def test_nested_group_yields_all_leaves(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(1., 0., 0.))
        t1 = QTrap(r=(0., 0., 0.), phase=0.)
        t2 = QTrap(r=(2., 0., 0.), phase=0.)
        inner.addTrap([t1, t2])
        t3 = QTrap(r=(5., 0., 0.), phase=0.)
        outer.addTrap([inner, t3])
        self.assertEqual(set(outer.leaves()), {t1, t2, t3})

    def test_leaves_are_not_groups(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(1., 0., 0.))
        inner.addTrap([QTrap(phase=0.), QTrap(phase=0.)])
        outer.addTrap(inner)
        for leaf in outer.leaves():
            self.assertNotIsInstance(leaf, QTrapGroup)

    def test_iter_yields_direct_children(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(1., 0., 0.))
        trap = QTrap(phase=0.)
        inner.addTrap(trap)
        outer.addTrap(inner)
        self.assertEqual(list(outer), [inner])

    def test_len_counts_direct_children(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(1., 0., 0.))
        inner.addTrap([QTrap(phase=0.), QTrap(phase=0.)])
        outer.addTrap(inner)
        self.assertEqual(len(outer), 1)


class TestRepr(unittest.TestCase):

    def test_repr(self):
        group = QTrapGroup(r=(0., 0., 0.))
        s = repr(group)
        self.assertTrue(s.startswith('QTrapGroup('))
        self.assertIn('ntraps=', s)


class TestGroupTranslation(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.))
        self.t1 = QTrap(r=(1., 0., 0.), phase=0.)
        self.t2 = QTrap(r=(2., 0., 0.), phase=0.)
        self.group.addTrap([self.t1, self.t2])

    def test_group_changed_emitted_once(self):
        spy = QtTest.QSignalSpy(self.group.changed)
        self.group.r = (3., 0., 0.)
        self.assertEqual(len(spy), 1)

    def test_leaf_changed_not_emitted_on_group_move(self):
        spy1 = QtTest.QSignalSpy(self.t1.changed)
        spy2 = QtTest.QSignalSpy(self.t2.changed)
        self.group.r = (3., 0., 0.)
        self.assertEqual(len(spy1), 0)
        self.assertEqual(len(spy2), 0)

    def test_leaf_positions_updated(self):
        self.group.r = (3., 0., 0.)
        np.testing.assert_array_almost_equal(self.t1.r, [4., 0., 0.])
        np.testing.assert_array_almost_equal(self.t2.r, [5., 0., 0.])

    def test_nested_group_positions_updated(self):
        inner = QTrapGroup(r=(5., 0., 0.))
        leaf = QTrap(r=(6., 0., 0.), phase=0.)
        inner.addTrap(leaf)
        outer = QTrapGroup(r=(0., 0., 0.))
        outer.addTrap(inner)
        outer.r = (2., 0., 0.)
        np.testing.assert_array_almost_equal(inner.r, [7., 0., 0.])
        np.testing.assert_array_almost_equal(leaf.r, [8., 0., 0.])


class TestTranslateSilently(unittest.TestCase):

    def test_updates_group_position(self):
        group = QTrapGroup(r=(1., 0., 0.))
        group._translateSilently(np.array([2., 0., 0.]))
        np.testing.assert_array_almost_equal(group.r, [3., 0., 0.])

    def test_updates_leaf_positions(self):
        group = QTrapGroup(r=(0., 0., 0.))
        t = QTrap(r=(1., 0., 0.), phase=0.)
        group.addTrap(t)
        group._translateSilently(np.array([5., 0., 0.]))
        np.testing.assert_array_almost_equal(t.r, [6., 0., 0.])

    def test_does_not_emit_leaf_changed(self):
        group = QTrapGroup(r=(0., 0., 0.))
        t = QTrap(r=(1., 0., 0.), phase=0.)
        group.addTrap(t)
        spy = QtTest.QSignalSpy(t.changed)
        group._translateSilently(np.array([1., 0., 0.]))
        self.assertEqual(len(spy), 0)

    def test_does_not_emit_group_changed(self):
        group = QTrapGroup(r=(0., 0., 0.))
        spy = QtTest.QSignalSpy(group.changed)
        group._translateSilently(np.array([1., 0., 0.]))
        self.assertEqual(len(spy), 0)


class TestSnapshot(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.))
        self.t1 = QTrap(r=(1., 0., 0.), phase=0.)
        self.t2 = QTrap(r=(2., 0., 0.), phase=0.)
        self.group.addTrap([self.t1, self.t2])

    def test_contains_all_children(self):
        snap = self.group._snapshot()
        self.assertIn(id(self.t1), snap)
        self.assertIn(id(self.t2), snap)

    def test_values_are_copies(self):
        snap = self.group._snapshot()
        self.t1._r[0] = 99.
        np.testing.assert_array_almost_equal(snap[id(self.t1)], [1., 0., 0.])

    def test_nested_children_included(self):
        inner = QTrapGroup(r=(5., 0., 0.))
        leaf = QTrap(r=(6., 0., 0.), phase=0.)
        inner.addTrap(leaf)
        outer = QTrapGroup(r=(0., 0., 0.))
        outer.addTrap(inner)
        snap = outer._snapshot()
        self.assertIn(id(inner), snap)
        self.assertIn(id(leaf), snap)


class TestRotate(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.))
        self.t1 = QTrap(r=(1., 0., 0.), phase=0.)
        self.t2 = QTrap(r=(-1., 0., 0.), phase=0.)
        self.group.addTrap([self.t1, self.t2])
        self.snap = self.group._snapshot()

    def test_90deg_rotation(self):
        self.group.rotate(np.pi / 2., self.snap)
        np.testing.assert_array_almost_equal(self.t1._r[:2], [0., 1.])
        np.testing.assert_array_almost_equal(self.t2._r[:2], [0., -1.])

    def test_180deg_rotation(self):
        self.group.rotate(np.pi, self.snap)
        np.testing.assert_array_almost_equal(self.t1._r[:2], [-1., 0.])
        np.testing.assert_array_almost_equal(self.t2._r[:2], [1., 0.])

    def test_group_center_unchanged(self):
        self.group.rotate(np.pi / 4., self.snap)
        np.testing.assert_array_almost_equal(self.group._r, [0., 0., 0.])

    def test_z_coordinate_unchanged(self):
        self.t1._r[2] = 5.
        snap = self.group._snapshot()
        self.group.rotate(np.pi / 2., snap)
        self.assertAlmostEqual(self.t1._r[2], 5.)

    def test_group_changed_emitted(self):
        spy = QtTest.QSignalSpy(self.group.changed)
        self.group.rotate(np.pi / 2., self.snap)
        self.assertGreaterEqual(len(spy), 1)

    def test_leaf_changed_emitted_for_cgh_invalidation(self):
        spy1 = QtTest.QSignalSpy(self.t1.changed)
        spy2 = QtTest.QSignalSpy(self.t2.changed)
        self.group.rotate(np.pi / 2., self.snap)
        self.assertGreaterEqual(len(spy1), 1)
        self.assertGreaterEqual(len(spy2), 1)

    def test_idempotent_with_same_snapshot(self):
        self.group.rotate(np.pi / 3., self.snap)
        r1a = self.t1._r.copy()
        self.group.rotate(np.pi / 3., self.snap)
        np.testing.assert_array_almost_equal(self.t1._r, r1a)

    def test_zero_angle_no_change(self):
        self.group.rotate(0., self.snap)
        np.testing.assert_array_almost_equal(self.t1._r, [1., 0., 0.])

    def test_non_zero_center(self):
        group = QTrapGroup(r=(2., 0., 0.))
        trap = QTrap(r=(3., 0., 0.), phase=0.)
        group.addTrap(trap)
        snap = group._snapshot()
        group.rotate(np.pi / 2., snap)
        np.testing.assert_array_almost_equal(trap._r[:2], [2., 1.])

    def test_nested_group_center_rotates(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(2., 0., 0.))
        leaf = QTrap(r=(3., 0., 0.), phase=0.)
        inner.addTrap(leaf)
        outer.addTrap(inner)
        snap = outer._snapshot()
        outer.rotate(np.pi / 2., snap)
        np.testing.assert_array_almost_equal(inner._r[:2], [0., 2.])
        np.testing.assert_array_almost_equal(leaf._r[:2], [0., 3.])

    def test_inner_group_changed_emitted_for_cgh_invalidation(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(2., 0., 0.))
        leaf = QTrap(r=(3., 0., 0.), phase=0.)
        inner.addTrap(leaf)
        outer.addTrap(inner)
        snap = outer._snapshot()
        spy_inner = QtTest.QSignalSpy(inner.changed)
        spy_leaf = QtTest.QSignalSpy(leaf.changed)
        outer.rotate(np.pi / 2., snap)
        self.assertGreaterEqual(len(spy_inner), 1)
        self.assertGreaterEqual(len(spy_leaf), 1)

    def test_nested_leaf_rotates_around_outer_center(self):
        outer = QTrapGroup(r=(0., 0., 0.))
        inner = QTrapGroup(r=(1., 0., 0.))
        leaf = QTrap(r=(2., 0., 0.), phase=0.)
        inner.addTrap(leaf)
        outer.addTrap(inner)
        snap = outer._snapshot()
        outer.rotate(np.pi / 2., snap)
        np.testing.assert_array_almost_equal(leaf._r[:2], [0., 2.])


class TestToDict(unittest.TestCase):

    def setUp(self):
        self.group = QTrapGroup(r=(0., 0., 0.), phase=0.)
        self.child_a = QTrap(r=(1., 2., 0.), phase=0.)
        self.child_b = QTrap(r=(3., 4., 0.), phase=0.)
        self.group.addTrap([self.child_a, self.child_b])

    def test_type_key(self):
        self.assertEqual(self.group.to_dict()['type'], 'QTrapGroup')

    def test_children_present(self):
        self.assertIn('children', self.group.to_dict())

    def test_children_count(self):
        self.assertEqual(len(self.group.to_dict()['children']), 2)

    def test_children_have_type(self):
        for child in self.group.to_dict()['children']:
            self.assertIn('type', child)

    def test_empty_group(self):
        empty = QTrapGroup(r=(0., 0., 0.), phase=0.)
        self.assertEqual(empty.to_dict()['children'], [])


if __name__ == '__main__':
    unittest.main()



'''Unit tests for Repeat.'''
import unittest

from pyqtgraph.Qt import QtCore, QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.lib.tasks.QTaskManager import QTaskManager
from QHOT.tasks.BeginRepeat import BeginRepeat
from QHOT.tasks.Delay import Delay
from QHOT.tasks.Repeat import Repeat

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class MockScreen(QtCore.QObject):
    rendered = QtCore.pyqtSignal()


# ------------------------------------------------------------------
# Init

class TestRepeatInit(unittest.TestCase):

    def test_registered_in_registry(self):
        self.assertIn('Repeat', QTask._registry)

    def test_duration_is_zero(self):
        self.assertEqual(Repeat().duration, 0)

    def test_default_n(self):
        self.assertEqual(Repeat().n, 1)

    def test_explicit_n_stored(self):
        self.assertEqual(Repeat(n=5).n, 5)

    def test_duration_keyword_raises(self):
        with self.assertRaises(TypeError):
            Repeat(duration=1)

    def test_parameters_declared(self):
        names = [p['name'] for p in Repeat.parameters]
        self.assertIn('n', names)

    def test_initial_state_pending(self):
        self.assertEqual(Repeat().state, QTask.State.PENDING)


# ------------------------------------------------------------------
# Serialisation

class TestRepeatSerialization(unittest.TestCase):

    def test_to_dict_includes_type(self):
        self.assertEqual(Repeat(n=3).to_dict()['type'], 'Repeat')

    def test_to_dict_includes_n(self):
        self.assertEqual(Repeat(n=3).to_dict()['n'], 3)

    def test_to_dict_no_specs_when_none(self):
        self.assertNotIn('_specs', Repeat(n=3).to_dict())

    def test_to_dict_includes_specs_when_set(self):
        specs = [{'type': 'Delay', 'delay': 0, 'frames': 0}]
        repeat = Repeat(n=2, _specs=specs)
        d = repeat.to_dict()
        self.assertIn('_specs', d)
        self.assertEqual(d['_specs'], specs)

    def test_round_trip_n(self):
        task = Repeat(n=4)
        restored = QTask.from_dict(task.to_dict())
        self.assertIsInstance(restored, Repeat)
        self.assertEqual(restored.n, 4)

    def test_round_trip_with_specs(self):
        specs = [{'type': 'Delay', 'delay': 0, 'frames': 0}]
        task = Repeat(n=2, _specs=specs)
        restored = QTask.from_dict(task.to_dict())
        self.assertIsInstance(restored, Repeat)
        self.assertEqual(restored._specs, specs)


# ------------------------------------------------------------------
# Execution helpers

class TestRepeatSetup(unittest.TestCase):
    '''Base class providing a manager wired to a MockScreen.'''

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _pump(self, max_signals=200):
        '''Fire rendered until manager pauses (schedule complete).

        Returns the number of signals fired.
        '''
        for count in range(1, max_signals + 1):
            self.screen.rendered.emit()
            if self.manager.paused:
                return count
        return max_signals  # did not complete


# ------------------------------------------------------------------
# Execution: no-op / single pass

class TestRepeatNoOp(TestRepeatSetup):

    def test_n1_emits_finished(self):
        repeat = Repeat(n=1)
        self.manager.register(repeat)
        spy = QtTest.QSignalSpy(repeat.finished)
        self.screen.rendered.emit()
        self.assertEqual(len(spy), 1)

    def test_n1_without_bracket_completes(self):
        '''Repeat with no BeginRepeat in schedule: logs warning, completes.'''
        self.manager.register(Repeat(n=3))
        self._pump()
        self.assertTrue(self.manager.paused)

    def test_manager_is_set_after_register(self):
        repeat = Repeat(n=1)
        self.manager.register(repeat)
        self.assertIs(repeat.manager, self.manager)


# ------------------------------------------------------------------
# Execution: bracketed block

class TestRepeatBracketed(TestRepeatSetup):

    def _schedule_block(self, n):
        '''Register [BeginRepeat, Delay(0), Repeat(n)] and return Repeat.'''
        self.manager.register(BeginRepeat())
        self.manager.register(Delay(frames=0))
        repeat = Repeat(n=n)
        self.manager.register(repeat)
        return repeat

    def test_n1_block_runs_once(self):
        # With n=1: no injection.  3 tasks × 1 step each = 3 rendered signals.
        self._schedule_block(n=1)
        count = self._pump()
        self.assertEqual(count, 3)

    def test_n2_block_runs_twice(self):
        # n=2: 3 (first pass) + 2 (Delay_fresh + Repeat(n=1)) = 5
        self._schedule_block(n=2)
        count = self._pump()
        self.assertEqual(count, 5)

    def test_n3_block_runs_three_times(self):
        # n=3: 3 + 2 + 2 = 7
        self._schedule_block(n=3)
        count = self._pump()
        self.assertEqual(count, 7)

    def test_rendered_count_formula(self):
        # General: 2*n + 1 rendered signals for a single-task block
        for n in range(1, 6):
            self.setUp()   # fresh manager each iteration
            self._schedule_block(n=n)
            count = self._pump()
            self.assertEqual(count, 2 * n + 1,
                             msg=f'n={n}: expected {2*n+1}, got {count}')

    def test_schedule_unchanged_after_block(self):
        '''Injected tasks must not appear in the persistent schedule.'''
        self._schedule_block(n=3)
        self._pump()
        self.assertEqual(len(self.manager.scheduled), 3)

    def test_manager_resets_after_completion(self):
        self._schedule_block(n=2)
        self._pump()
        self.assertTrue(self.manager.paused)


# ------------------------------------------------------------------
# Execution: nested brackets

class TestRepeatNested(TestRepeatSetup):

    def test_nested_outer_runs_correct_count(self):
        # [BeginRepeat, BeginRepeat, Delay(0), Repeat(n=2), Repeat(n=3)]
        # Inner block: Delay runs 2× per outer pass.
        # Outer: 3 outer passes.
        # Rendered: outer-BeginRepeat(1) + 3×(inner-pass)
        # inner-pass = inner-BeginRepeat(1) + Delay(1) + inner-Repeat(1) = 3
        # But on outer runs 2 & 3, outer-BeginRepeat is not re-run (it's fresh).
        # Let me count carefully:
        # Pass 1 (from schedule): BR_outer + BR_inner + Delay + Repeat(n=2) +
        #   inject [BR_inner_f, Delay_f, Repeat(n=1)] + outer_Repeat(n=3) runs
        # This is complex to predict exactly; just verify completion.
        self.manager.register(BeginRepeat())   # outer
        self.manager.register(BeginRepeat())   # inner
        self.manager.register(Delay(frames=0))
        self.manager.register(Repeat(n=2))     # inner repeat
        self.manager.register(Repeat(n=3))     # outer repeat
        count = self._pump(max_signals=500)
        self.assertTrue(self.manager.paused,
                        f'did not complete within {count} signals')
        self.assertEqual(len(self.manager.scheduled), 5)

    def test_inner_specs_do_not_include_outer_tasks(self):
        '''Inner Repeat should only capture tasks between its BeginRepeat.'''
        self.manager.register(BeginRepeat())   # outer
        self.manager.register(Delay(frames=0))  # outer-only task
        self.manager.register(BeginRepeat())   # inner
        self.manager.register(Delay(frames=0))  # inner task
        inner_repeat = Repeat(n=2)
        self.manager.register(inner_repeat)
        self.manager.register(Repeat(n=1))     # outer no-op

        # Step through until inner_repeat has initialised
        for _ in range(10):
            self.screen.rendered.emit()
            if inner_repeat.state == QTask.State.COMPLETED:
                break

        # _specs should contain only the inner Delay, not the outer one
        self.assertIsNotNone(inner_repeat._specs)
        self.assertEqual(len(inner_repeat._specs), 1)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



'''Unit tests for SaveTraps.'''
import unittest
from unittest.mock import MagicMock

from pyqtgraph.Qt import QtWidgets

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.SaveTraps import SaveTraps

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


def _run(task):
    task._start()
    task._step()


class TestSaveTrapsInit(unittest.TestCase):

    def test_registered_in_registry(self):
        self.assertIn('SaveTraps', QTask._registry)

    def test_duration_is_zero(self):
        self.assertEqual(SaveTraps().duration, 0)

    def test_default_filename_empty(self):
        self.assertEqual(SaveTraps().filename, '')

    def test_explicit_filename_stored(self):
        task = SaveTraps(filename='/tmp/traps.json')
        self.assertEqual(task.filename, '/tmp/traps.json')

    def test_parameters_declared(self):
        names = [p['name'] for p in SaveTraps.parameters]
        self.assertIn('filename', names)


class TestSaveTrapsExecution(unittest.TestCase):

    def setUp(self):
        self.overlay = MagicMock()
        self.save    = MagicMock()

    def test_save_traps_called_with_overlay(self):
        task = SaveTraps(overlay=self.overlay, save=self.save)
        _run(task)
        self.save.traps.assert_called_once()
        args, _ = self.save.traps.call_args
        self.assertIs(args[0], self.overlay)

    def test_explicit_filename_passed(self):
        task = SaveTraps(overlay=self.overlay, save=self.save,
                         filename='/tmp/traps.json')
        _run(task)
        _, kwargs = self.save.traps.call_args
        self.assertEqual(kwargs['filename'], '/tmp/traps.json')

    def test_empty_filename_passes_none(self):
        task = SaveTraps(overlay=self.overlay, save=self.save, filename='')
        _run(task)
        _, kwargs = self.save.traps.call_args
        self.assertIsNone(kwargs['filename'])

    def test_completes_after_one_step(self):
        task = SaveTraps(overlay=self.overlay, save=self.save)
        _run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)


class TestSaveTrapsSerialization(unittest.TestCase):

    def test_to_dict_includes_type(self):
        self.assertEqual(SaveTraps().to_dict()['type'], 'SaveTraps')

    def test_round_trip(self):
        task = SaveTraps(filename='/tmp/traps.json')
        restored = QTask.from_dict(task.to_dict())
        self.assertIsInstance(restored, SaveTraps)
        self.assertEqual(restored.filename, '/tmp/traps.json')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



'''Unit tests for QTaskManagerWidget.'''
import unittest
from unittest.mock import MagicMock

from pyqtgraph.Qt import QtCore, QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.lib.tasks.QTaskManager import QTaskManager
from QHOT.lib.tasks.QTaskManagerWidget import QTaskManagerWidget, _ROLE

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class MockScreen(QtCore.QObject):
    rendered = QtCore.pyqtSignal()


def _make_manager():
    screen = MockScreen()
    return QTaskManager(screen)


def _make_manager_with_screen():
    screen = MockScreen()
    return QTaskManager(screen), screen


def _make_wired():
    screen = MockScreen()
    manager = QTaskManager(screen)
    widget = QTaskManagerWidget()
    widget.manager = manager
    return screen, manager, widget


class TestQTaskManagerWidgetInit(unittest.TestCase):

    def setUp(self):
        self.widget = QTaskManagerWidget()

    def test_manager_is_none_initially(self):
        self.assertIsNone(self.widget.manager)

    def test_play_button_disabled_without_manager(self):
        self.assertFalse(self.widget._playButton.isEnabled())

    def test_stop_button_disabled_without_manager(self):
        self.assertFalse(self.widget._stopButton.isEnabled())

    def test_clear_button_disabled_without_manager(self):
        self.assertFalse(self.widget._clearButton.isEnabled())

    def test_status_emitted_on_init(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.widget._refresh()
        self.assertEqual(len(spy), 1)
        self.assertIn('not connected', spy[0][0])

    def test_queue_list_empty(self):
        self.assertEqual(self.widget._queueList.count(), 0)

    def test_background_list_empty(self):
        self.assertEqual(self.widget._bgList.count(), 0)


class TestQTaskManagerWidgetManagerProperty(unittest.TestCase):

    def setUp(self):
        self.widget  = QTaskManagerWidget()
        self.manager = _make_manager()

    def test_setting_manager_stores_it(self):
        self.widget.manager = self.manager
        self.assertIs(self.widget.manager, self.manager)

    def test_setting_manager_enables_clear_button(self):
        self.widget.manager = self.manager
        self.assertTrue(self.widget._clearButton.isEnabled())

    def test_play_stop_disabled_when_schedule_empty(self):
        self.widget.manager = self.manager
        self.assertFalse(self.widget._playButton.isEnabled())
        self.assertFalse(self.widget._stopButton.isEnabled())

    def test_play_stop_enabled_after_task_registered(self):
        self.widget.manager = self.manager
        self.manager.register(QTask())
        self.assertTrue(self.widget._playButton.isEnabled())
        self.assertTrue(self.widget._stopButton.isEnabled())

    def test_setting_none_disables_buttons(self):
        self.widget.manager = self.manager
        self.manager.register(QTask())
        self.widget.manager = None
        self.assertFalse(self.widget._playButton.isEnabled())
        self.assertFalse(self.widget._stopButton.isEnabled())
        self.assertFalse(self.widget._clearButton.isEnabled())

    def test_same_manager_is_noop(self):
        self.widget.manager = self.manager
        # Should not raise or double-connect
        self.widget.manager = self.manager

    def test_replacing_manager_disconnects_old(self):
        old = _make_manager()
        self.widget.manager = old
        spy = QtTest.QSignalSpy(old.changed)
        self.widget.manager = self.manager
        old.pause(True)              # old.changed fires
        self.assertEqual(len(spy), 1)
        # widget should NOT have refreshed from old manager's signal
        # (just verifying no error is raised)

    def test_setting_manager_triggers_refresh(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.widget.manager = self.manager
        self.assertGreater(len(spy), 0)

    def test_status_idle_when_manager_set_empty(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.widget.manager = self.manager
        self.assertTrue(any('Idle' in s[0] for s in spy))


class TestQTaskManagerWidgetDisplay(unittest.TestCase):

    def setUp(self):
        self.widget  = QTaskManagerWidget()
        self.manager, self.screen = _make_manager_with_screen()
        self.widget.manager = self.manager

    def _emit(self, n=1):
        for _ in range(n):
            self.screen.rendered.emit()

    def test_registered_task_appears_in_queue_list(self):
        task = QTask()
        self.manager.register(task)
        self.assertEqual(self.widget._queueList.count(), 1)
        self.assertEqual(self.widget._queueList.item(0).text(), 'QTask')

    def test_active_task_shown_bold(self):
        task = QTask()
        self.manager.register(task)
        self._emit()          # first frame: task becomes active
        item = self.widget._queueList.item(0)
        self.assertTrue(item.font().bold())

    def test_active_task_has_nyu_tint_background(self):
        from pyqtgraph.Qt import QtGui
        task = QTask()
        self.manager.register(task)
        self._emit()
        item = self.widget._queueList.item(0)
        self.assertEqual(item.background().color().name(), '#e8d5f5')

    def test_completed_task_remains_in_queue_list(self):
        task = QTask(duration=1)
        self.manager.register(task)
        self._emit(1)         # task completes
        self.assertEqual(self.widget._queueList.count(), 1)

    def test_completed_task_shown_gray(self):
        t1 = QTask(duration=1)
        t2 = QTask()          # won't auto-complete; keeps t1 visible as COMPLETED
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit(1)         # t1 completes, t2 activates
        item = self.widget._queueList.item(0)
        color = item.foreground().color()
        self.assertEqual(color.red(),   128)
        self.assertEqual(color.green(), 128)
        self.assertEqual(color.blue(),  128)

    def test_failed_task_shown_red(self):
        task = QTask()
        self.manager.register(task)
        task.abort('test')
        item = self.widget._queueList.item(0)
        color = item.foreground().color()
        self.assertEqual(color.red(),   192)
        self.assertEqual(color.green(),   0)
        self.assertEqual(color.blue(),    0)

    def test_pending_task_not_bold(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit()          # first frame: t1 active (index 0), t2 pending (index 1)
        item = self.widget._queueList.item(1)
        self.assertFalse(item.font().bold())

    def test_multiple_tasks_all_shown(self):
        for _ in range(3):
            self.manager.register(QTask())
        self.assertEqual(self.widget._queueList.count(), 3)

    def test_queue_list_count_stable_during_execution(self):
        t1 = QTask(duration=1)
        t2 = QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit(1)         # t1 completes, t2 activates
        self._emit(1)         # t2 stepped
        self.assertEqual(self.widget._queueList.count(), 2)

    def test_background_list_shows_task_name(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        self.assertEqual(self.widget._bgList.count(), 1)
        self.assertEqual(self.widget._bgList.item(0).text(), 'QTask')

    def test_background_list_clears_after_task_finishes(self):
        task = QTask(duration=1)
        self.manager.register(task, blocking=False)
        task._step()
        self.assertEqual(self.widget._bgList.count(), 0)

    def test_status_running_when_active_task(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.manager.register(QTask())
        self._emit()          # first frame: task becomes active
        self.assertTrue(any('Running' in s[0] for s in spy))

    def test_status_running_with_background_only(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.manager.register(QTask(), blocking=False)
        self.assertTrue(any('Running' in s[0] for s in spy))

    def test_status_idle_when_nothing_running(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.widget._refresh()
        self.assertTrue(any('Idle' in s[0] for s in spy))


class TestQTaskManagerWidgetControls(unittest.TestCase):

    def setUp(self):
        self.widget  = QTaskManagerWidget()
        self.manager, self.screen = _make_manager_with_screen()
        self.widget.manager = self.manager

    def _emit(self, n=1):
        for _ in range(n):
            self.screen.rendered.emit()

    def test_play_button_shows_play_when_idle(self):
        self.assertEqual(self.widget._playButton.text(), 'Play')

    def test_play_button_shows_pause_when_running(self):
        self.manager.register(QTask())
        self._emit()          # first frame: task becomes active
        self.assertEqual(self.widget._playButton.text(), 'Pause')

    def test_play_button_shows_play_when_paused(self):
        self.manager.register(QTask())
        self.manager.pause(True)
        self.assertEqual(self.widget._playButton.text(), 'Play')

    def test_clicking_play_unpauses_manager(self):
        self.manager.register(QTask())
        self.manager.pause(True)
        self.widget._playButton.click()
        self.assertFalse(self.manager.paused)

    def test_clicking_pause_pauses_manager(self):
        self.manager.register(QTask())
        self._emit()          # task is active (not paused)
        self.widget._playButton.click()
        self.assertTrue(self.manager.paused)

    def test_status_paused_when_manager_paused(self):
        spy = QtTest.QSignalSpy(self.widget.status)
        self.manager.pause(True)
        self.assertTrue(any('Paused' in s[0] for s in spy))

    def test_stop_rewinds_to_start(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit()          # t1 active
        self.widget._stopButton.click()
        # After rewind: paused at first task, not yet stepped
        self.assertIsNone(self.manager.active)
        self.assertTrue(self.manager.paused)

    def test_stop_preserves_schedule(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.widget._stopButton.click()
        self.assertEqual(self.widget._queueList.count(), 2)

    def test_clear_button_empties_queue_list(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.widget._clearButton.click()
        self.assertEqual(self.widget._queueList.count(), 0)

    def test_clear_button_clears_background_list(self):
        self.manager.register(QTask(), blocking=False)
        self.widget._clearButton.click()
        self.assertEqual(self.widget._bgList.count(), 0)

    def test_play_button_no_op_without_manager(self):
        self.widget.manager = None
        self.widget._onPlayClicked()    # should not raise

    def test_stop_button_no_op_without_manager(self):
        self.widget.manager = None
        self.widget._onStopClicked()    # should not raise

    def test_clear_button_no_op_without_manager(self):
        self.widget.manager = None
        self.widget._onClearClicked()   # should not raise


class TestQTaskManagerWidgetParamTree(unittest.TestCase):

    def setUp(self):
        self.widget  = QTaskManagerWidget()
        self.manager, self.screen = _make_manager_with_screen()
        self.widget.manager = self.manager

    def _emit(self, n=1):
        for _ in range(n):
            self.screen.rendered.emit()

    def test_param_tree_empty_initially(self):
        self.assertIsNone(self.widget._taskTree)

    def test_clicking_active_task_populates_param_tree(self):
        from QHOT.tasks.Delay import Delay
        task = Delay(frames=100)
        self.manager.register(task)
        self._emit()          # first frame: task becomes active (index 0)
        self.manager.pause(True)
        item = self.widget._queueList.item(0)
        self.widget._onTaskItemClicked(item)
        self.assertIsNotNone(self.widget._taskTree)
        root = self.widget._taskTree.invisibleRootItem()
        self.assertEqual(root.childCount(), 1)
        n_params = len(type(task).parameters)
        self.assertEqual(root.child(0).childCount(), n_params)

    def test_clicking_pending_task_populates_param_tree(self):
        from QHOT.tasks.Delay import Delay
        t1 = Delay(frames=100)
        t2 = Delay(frames=200)
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit()          # first frame: t1 active (index 0), t2 pending (index 1)
        item = self.widget._queueList.item(1)
        self.widget._onTaskItemClicked(item)
        self.assertIsNotNone(self.widget._taskTree)
        root = self.widget._taskTree.invisibleRootItem()
        self.assertEqual(root.childCount(), 1)
        n_params = len(type(t2).parameters)
        self.assertEqual(root.child(0).childCount(), n_params)

    def test_param_tree_removed_when_task_cleared(self):
        from QHOT.tasks.Delay import Delay
        task = Delay(frames=100)
        self.manager.register(task)
        self._emit()
        self.manager.pause(True)
        item = self.widget._queueList.item(0)
        self.widget._onTaskItemClicked(item)
        # Clear removes all tasks; _reselectTask should remove the tree
        self.manager.clear()
        self.assertIsNone(self.widget._selectedTask)
        self.assertIsNone(self.widget._taskTree)

    def test_selected_task_reselected_after_refresh(self):
        t1 = QTask()
        t2 = QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit()          # first frame: t1 active (index 0), t2 pending (index 1)
        item = self.widget._queueList.item(1)
        self.widget._onTaskItemClicked(item)
        self.assertIs(self.widget._selectedTask, t2)
        self.manager.pause(True)
        self.assertIs(self.widget._selectedTask, t2)


class TestQTaskManagerWidgetRemove(unittest.TestCase):

    def setUp(self):
        self.screen, self.manager, self.widget = _make_wired()

    def test_delete_key_removes_selected_pending_task(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.widget._queueList.setCurrentRow(1)   # select t2
        self.widget._onDeletePressed()
        self.assertNotIn(t2, self.manager.scheduled)

    def test_delete_key_does_not_remove_active_task(self):
        t1 = QTask()
        self.manager.register(t1)
        self.widget._queueList.setCurrentRow(0)   # select t1 (running)
        self.widget._onDeletePressed()
        self.assertIn(t1, self.manager.scheduled)

    def test_delete_key_no_selection_is_noop(self):
        t1 = QTask()
        self.manager.register(t1)
        self.widget._queueList.clearSelection()
        self.widget._onDeletePressed()   # should not raise
        self.assertIn(t1, self.manager.scheduled)

    def test_delete_key_no_manager_is_noop(self):
        self.widget.manager = None
        self.widget._onDeletePressed()   # should not raise

    def test_context_menu_remove_action_calls_manager_remove(self):
        import sys
        from unittest.mock import patch, MagicMock
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        item = self.widget._queueList.item(1)
        fake_menu = MagicMock()
        remove_action = MagicMock()
        fake_menu.addAction.return_value = remove_action
        fake_menu.exec.return_value = remove_action
        module = sys.modules['QHOT.lib.tasks.QTaskManagerWidget']
        with patch.object(module, 'QtWidgets') as mock_qtw:
            mock_qtw.QMenu.return_value = fake_menu
            with patch.object(self.manager, 'remove') as mock_remove:
                self.widget._onQueueContextMenu(
                    self.widget._queueList.visualItemRect(item).center())
                mock_remove.assert_called_once_with(t2)

    def test_context_menu_remove_disabled_for_active_task(self):
        import sys
        from unittest.mock import patch, MagicMock
        t1 = QTask()
        self.manager.register(t1)
        item = self.widget._queueList.item(0)
        fake_menu = MagicMock()
        remove_action = MagicMock()
        fake_menu.addAction.return_value = remove_action
        fake_menu.exec.return_value = None
        module = sys.modules['QHOT.lib.tasks.QTaskManagerWidget']
        with patch.object(module, 'QtWidgets') as mock_qtw:
            mock_qtw.QMenu.return_value = fake_menu
            self.widget._onQueueContextMenu(
                self.widget._queueList.visualItemRect(item).center())
            remove_action.setEnabled.assert_called_once_with(False)

    def test_context_menu_no_item_is_noop(self):
        self.widget._onQueueContextMenu(QtCore.QPoint(0, 0))  # empty list area


class TestDraggableQueueListCanDropAt(unittest.TestCase):
    '''Tests for _DraggableQueueList._canDropAt().

    Row 0 is RUNNING; rows 1 and 2 are PENDING (first_pending=1).
    Inserts below row 0 are allowed; inserts at row 0 or above are not.

    _canDropAt is pure Python so no Qt event objects are needed.
    '''

    def setUp(self):
        from QHOT.lib.tasks.QTaskManagerWidget import _DraggableQueueList
        screen = MockScreen()
        manager = QTaskManager(screen)
        t1, t2, t3 = QTask(), QTask(), QTask()
        manager.register(t1)   # RUNNING at row 0
        manager.register(t2)   # PENDING at row 1
        manager.register(t3)   # PENDING at row 2
        self.lst = _DraggableQueueList()
        for task in (t1, t2, t3):
            item = QtWidgets.QListWidgetItem()
            item.setData(_ROLE, task)
            self.lst.addItem(item)

    def test_insert_before_running_rejected(self):
        # insert_row=0 < first_pending=1 → rejected
        self.assertFalse(self.lst._canDropAt(0))

    def test_insert_at_first_pending_allowed(self):
        # insert_row=1 == first_pending=1 → allowed
        # This is the key fix: old code treated row-under-cursor as the
        # insert row, which returned 0 (RUNNING) when hovering near the
        # boundary above row 1, incorrectly rejecting the drop.
        self.assertTrue(self.lst._canDropAt(1))

    def test_insert_between_pending_allowed(self):
        # insert_row=2 > first_pending=1 → allowed
        self.assertTrue(self.lst._canDropAt(2))

    def test_insert_past_end_allowed(self):
        # insert_row=count=3 > first_pending=1 → allowed
        self.assertTrue(self.lst._canDropAt(self.lst.count()))

    def test_all_pending_any_insert_allowed(self):
        # When no RUNNING/COMPLETED tasks exist, first_pending=0 → always allowed
        from QHOT.lib.tasks.QTaskManagerWidget import _DraggableQueueList
        lst = _DraggableQueueList()
        tasks = [QTask() for _ in range(3)]  # keep alive: setData holds only a Qt-level ref
        for task in tasks:
            item = QtWidgets.QListWidgetItem()
            item.setData(_ROLE, task)
            lst.addItem(item)
        self.assertTrue(lst._canDropAt(0))


class TestQTaskManagerWidgetDragReorder(unittest.TestCase):

    def setUp(self):
        self.screen, self.manager, self.widget = _make_wired()

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_queue_list_has_internal_move(self):
        from pyqtgraph.Qt import QtWidgets as _QtW
        self.assertEqual(
            self.widget._queueList.dragDropMode(),
            _QtW.QAbstractItemView.DragDropMode.InternalMove)

    def test_pending_items_are_draggable(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        # t1 is active (RUNNING), t2 is PENDING
        item_t2 = self.widget._queueList.item(1)
        drag_flag = QtCore.Qt.ItemFlag.ItemIsDragEnabled
        self.assertTrue(bool(item_t2.flags() & drag_flag))

    def test_running_item_is_not_draggable(self):
        t1 = QTask()
        self.manager.register(t1)
        item_t1 = self.widget._queueList.item(0)
        drag_flag = QtCore.Qt.ItemFlag.ItemIsDragEnabled
        self.assertFalse(bool(item_t1.flags() & drag_flag))

    def test_on_rows_moved_calls_reorder(self):
        from unittest.mock import MagicMock, patch
        t1, t2, t3 = QTask(), QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.manager.register(t3)
        with patch.object(self.manager, 'reorder') as mock_reorder:
            # Simulate rowsMoved: pretend t3 (row 2) moved to row 1
            self.widget._onRowsMoved(
                QtCore.QModelIndex(), 2, 2, QtCore.QModelIndex(), 1)
            mock_reorder.assert_called_once()
            args = mock_reorder.call_args[0][0]
            self.assertEqual(len(args), 3)
            self.assertTrue(all(isinstance(t, QTask) for t in args))

    def test_on_rows_moved_no_manager_does_nothing(self):
        self.widget.manager = None
        # Should not raise
        self.widget._onRowsMoved(
            QtCore.QModelIndex(), 0, 0, QtCore.QModelIndex(), 1)

    def test_reorder_reflected_in_queue_list(self):
        t1, t2, t3 = QTask(), QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.manager.register(t3)
        self.manager.reorder([t1, t3, t2])
        items = [self.widget._queueList.item(i).data(_ROLE)
                 for i in range(self.widget._queueList.count())]
        self.assertEqual(items, [t1, t3, t2])


if __name__ == '__main__':
    unittest.main()



'''Unit tests for QueueMenu.'''
import unittest
from unittest.mock import MagicMock, patch

from pyqtgraph.Qt import QtCore, QtWidgets

import QHOT.tasks  # noqa: F401 — ensure registry is populated
from QHOT.lib.tasks.QTask import QTask
from QHOT.lib.tasks.QTaskManager import QTaskManager
from QHOT.lib.tasks.QueueMenu import QueueMenu

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class MockScreen(QtCore.QObject):
    rendered = QtCore.pyqtSignal()


def _make_manager():
    screen = MockScreen()
    return QTaskManager(screen)


class TestQueueMenuInit(unittest.TestCase):

    def setUp(self):
        self.menu = QueueMenu()

    def test_default_title(self):
        self.assertEqual(self.menu.title(), 'Queue')

    def test_custom_title(self):
        menu = QueueMenu(title='My Queue')
        self.assertEqual(menu.title(), 'My Queue')

    def test_manager_none_initially(self):
        self.assertIsNone(self.menu.manager)

    def test_overlay_none_initially(self):
        self.assertIsNone(self.menu.overlay)

    def test_cgh_none_initially(self):
        self.assertIsNone(self.menu.cgh)

    def test_dvr_none_initially(self):
        self.assertIsNone(self.menu.dvr)

    def test_actions_populated_from_registry(self):
        names = {a.text() for a in self.menu.actions()}
        for name in QTask._registry:
            self.assertIn(name, names)

    def test_has_at_least_one_action(self):
        self.assertGreater(len(self.menu.actions()), 0)


class TestQueueMenuProperties(unittest.TestCase):

    def setUp(self):
        self.menu = QueueMenu()

    def test_set_manager(self):
        mgr = _make_manager()
        self.menu.manager = mgr
        self.assertIs(self.menu.manager, mgr)

    def test_set_manager_to_none(self):
        self.menu.manager = _make_manager()
        self.menu.manager = None
        self.assertIsNone(self.menu.manager)

    def test_set_overlay(self):
        overlay = MagicMock()
        self.menu.overlay = overlay
        self.assertIs(self.menu.overlay, overlay)

    def test_set_cgh(self):
        cgh = MagicMock()
        self.menu.cgh = cgh
        self.assertIs(self.menu.cgh, cgh)

    def test_set_dvr(self):
        dvr = MagicMock()
        self.menu.dvr = dvr
        self.assertIs(self.menu.dvr, dvr)


class TestQueueMenuSelection(unittest.TestCase):

    def setUp(self):
        self.menu = QueueMenu()
        self.manager = _make_manager()

    def test_selecting_task_registers_with_manager(self):
        name = next(iter(QTask._registry))
        self.menu.manager = self.manager
        self.menu._onTaskSelected(name)
        # Either active_raw or in background — something was registered
        registered = (self.manager.active_raw is not None or
                      len(self.manager.background) > 0)
        self.assertTrue(registered)

    def test_no_manager_logs_warning(self):
        name = next(iter(QTask._registry))
        with self.assertLogs('QHOT.lib.tasks.QueueMenu', level='WARNING') as cm:
            self.menu._onTaskSelected(name)
        self.assertTrue(any('No manager' in line for line in cm.output))

    def test_unknown_task_logs_warning(self):
        self.menu.manager = self.manager
        with self.assertLogs('QHOT.lib.tasks.QueueMenu', level='WARNING') as cm:
            self.menu._onTaskSelected('NonExistentTask')
        self.assertTrue(any('Unknown task' in line for line in cm.output))

    def test_auto_pause_when_registering_into_idle_manager(self):
        self.menu.manager = self.manager
        name = next(iter(QTask._registry))
        self.menu._onTaskSelected(name)
        self.assertTrue(self.manager.paused)

    def test_no_auto_pause_when_manager_already_has_active_task(self):
        blocking = QTask()
        self.manager.register(blocking)
        self.menu.manager = self.manager
        name = next(iter(QTask._registry))
        self.menu._onTaskSelected(name)
        self.assertFalse(self.manager.paused)

    def test_no_auto_pause_when_background_task_running(self):
        bg = QTask()
        self.manager.register(bg, blocking=False)
        self.menu.manager = self.manager
        name = next(iter(QTask._registry))
        self.menu._onTaskSelected(name)
        self.assertFalse(self.manager.paused)

    def test_dependencies_injected_into_task(self):
        overlay = MagicMock()
        cgh = MagicMock()
        dvr = MagicMock()
        self.menu.manager = self.manager
        self.menu.overlay = overlay
        self.menu.cgh = cgh
        self.menu.dvr = dvr

        name = next(iter(QTask._registry))
        cls = QTask._registry[name]
        created = []
        original_init = cls.__init__

        def capture_init(self_task, *args, **kwargs):
            created.append(kwargs)
            original_init(self_task, *args, **kwargs)

        with patch.object(cls, '__init__', capture_init):
            self.menu._onTaskSelected(name)

        self.assertTrue(len(created) > 0)
        self.assertIs(created[0].get('overlay'), overlay)
        self.assertIs(created[0].get('cgh'), cgh)
        self.assertIs(created[0].get('dvr'), dvr)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for Delay.'''
import unittest
from unittest.mock import MagicMock

from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.Delay import Delay

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class TestDelayInit(unittest.TestCase):

    def test_duration_set_from_frames(self):
        task = Delay(10)
        self.assertEqual(task.duration, 10)

    def test_setting_frames_updates_duration(self):
        task = Delay(10)
        task.frames = 50
        self.assertEqual(task.duration, 50)

    def test_initial_state_is_pending(self):
        task = Delay(5)
        self.assertEqual(task.state, QTask.State.PENDING)

    def test_duration_kwarg_raises(self):
        with self.assertRaises(TypeError):
            Delay(5, duration=10)

    def test_zero_frames_accepted(self):
        task = Delay(0)
        self.assertEqual(task.duration, 0)


class TestDelayExecution(unittest.TestCase):

    def test_running_after_start(self):
        task = Delay(5)
        task._start()
        self.assertEqual(task.state, QTask.State.RUNNING)

    def test_still_running_before_duration_expires(self):
        task = Delay(5)
        task._start()
        for _ in range(4):
            task._step()
        self.assertEqual(task.state, QTask.State.RUNNING)

    def test_completes_after_exact_frames(self):
        task = Delay(5)
        task._start()
        for _ in range(5):
            task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_does_not_complete_one_step_early(self):
        task = Delay(10)
        task._start()
        for _ in range(9):
            task._step()
        self.assertNotEqual(task.state, QTask.State.COMPLETED)

    def test_process_never_overridden(self):
        # process() in Delay is the base no-op; verify it does nothing
        task = Delay(3)
        task._start()
        for _ in range(3):
            task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_delay_one_frame(self):
        task = Delay(1)
        task._start()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_zero_frame_delay_completes_immediately(self):
        task = Delay(0)
        task._start()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_finished_signal_emitted(self):
        task = Delay(2)
        spy = QtTest.QSignalSpy(task.finished)
        task._start()
        task._step()
        task._step()
        self.assertEqual(len(spy), 1)

    def test_finished_emitted_exactly_once(self):
        task = Delay(2)
        spy = QtTest.QSignalSpy(task.finished)
        task._start()
        for _ in range(5):
            task._step()
        self.assertEqual(len(spy), 1)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for lib/traps/commands.py.'''
import sys
import unittest
import numpy as np
from pyqtgraph.Qt import QtWidgets
from QHOT.lib.traps.QTrap import QTrap
from QHOT.lib.traps.QTrapGroup import QTrapGroup
from QHOT.lib.traps.QTrapOverlay import QTrapOverlay
from QHOT.lib.traps.commands import (
    AddTrapCommand, RemoveTrapCommand,
    MoveCommand, RotateCommand, WheelCommand,
    LockCommand, QUndoStack)
from QHOT.traps.QTweezer import QTweezer


app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)


def make_overlay():
    return QTrapOverlay()


class TestAddTrapCommand(unittest.TestCase):

    def setUp(self):
        self.overlay = make_overlay()
        self.stack = self.overlay._undoStack
        self.stack.push(AddTrapCommand(self.overlay, 3., 7.))

    def test_redo_adds_trap(self):
        self.assertEqual(len(self.overlay._traps), 1)

    def test_trap_is_tweezer(self):
        self.assertIsInstance(self.overlay._traps[0], QTweezer)

    def test_trap_position(self):
        trap = self.overlay._traps[0]
        self.assertAlmostEqual(trap.x, 3.)
        self.assertAlmostEqual(trap.y, 7.)

    def test_trap_added_signal(self):
        overlay = make_overlay()
        received = []
        overlay.trapAdded.connect(received.append)
        overlay._undoStack.push(AddTrapCommand(overlay, 1., 2.))
        self.assertEqual(len(received), 1)

    def test_undo_removes_trap(self):
        self.stack.undo()
        self.assertEqual(len(self.overlay._traps), 0)

    def test_undo_emits_trap_removed(self):
        received = []
        self.overlay.trapRemoved.connect(received.append)
        self.stack.undo()
        self.assertEqual(len(received), 1)

    def test_redo_after_undo_restores_trap(self):
        self.stack.undo()
        self.stack.redo()
        self.assertEqual(len(self.overlay._traps), 1)

    def test_repeated_undo_redo_cycle(self):
        for _ in range(3):
            self.stack.undo()
            self.assertEqual(len(self.overlay._traps), 0)
            self.stack.redo()
            self.assertEqual(len(self.overlay._traps), 1)


class TestRemoveTrapCommand(unittest.TestCase):

    def setUp(self):
        self.overlay = make_overlay()
        self.stack = self.overlay._undoStack
        self.trap = QTrap(r=(5., 5., 0.), phase=0.)
        self.overlay._addTrap(self.trap)
        self.stack.push(RemoveTrapCommand(self.overlay, self.trap))

    def test_redo_removes_trap(self):
        self.assertNotIn(self.trap, self.overlay._traps)

    def test_trap_removed_signal_emitted(self):
        overlay = make_overlay()
        trap = QTrap(r=(1., 1., 0.), phase=0.)
        overlay._addTrap(trap)
        received = []
        overlay.trapRemoved.connect(received.append)
        overlay._undoStack.push(RemoveTrapCommand(overlay, trap))
        self.assertEqual(len(received), 1)

    def test_undo_adds_trap_back(self):
        self.stack.undo()
        self.assertIn(self.trap, self.overlay._traps)

    def test_undo_emits_trap_added(self):
        received = []
        self.overlay.trapAdded.connect(received.append)
        self.stack.undo()
        self.assertEqual(len(received), 1)

    def test_redo_after_undo_removes_again(self):
        self.stack.undo()
        self.stack.redo()
        self.assertNotIn(self.trap, self.overlay._traps)

    def test_repeated_undo_redo_cycle(self):
        for _ in range(3):
            self.stack.undo()
            self.assertIn(self.trap, self.overlay._traps)
            self.stack.redo()
            self.assertNotIn(self.trap, self.overlay._traps)


class TestMoveCommand(unittest.TestCase):

    def setUp(self):
        self.trap = QTrap(r=(1., 2., 0.), phase=0.)
        self.origin = self.trap._r.copy()
        self.trap.r = (4., 5., 0.)
        self.cmd = MoveCommand(self.trap, self.origin)
        self.stack = self.trap._undoStack = QUndoStack()
        self.stack.push(self.cmd)

    def test_first_redo_is_noop(self):
        np.testing.assert_array_almost_equal(self.trap._r, [4., 5., 0.])

    def test_undo_restores_origin(self):
        self.stack.undo()
        np.testing.assert_array_almost_equal(self.trap._r, [1., 2., 0.])

    def test_redo_after_undo_restores_destination(self):
        self.stack.undo()
        self.stack.redo()
        np.testing.assert_array_almost_equal(self.trap._r, [4., 5., 0.])

    def test_undo_emits_changed(self):
        received = []
        self.trap.changed.connect(lambda: received.append(1))
        self.stack.undo()
        self.assertTrue(len(received) > 0)

    def test_text(self):
        self.assertEqual(self.cmd.text(), 'Move trap')


class TestRotateCommand(unittest.TestCase):

    def setUp(self):
        self.t1 = QTweezer(r=(4., 0., 0.))
        self.t2 = QTweezer(r=(-4., 0., 0.))
        self.group = QTrapGroup(r=(0., 0., 0.))
        self.group.addTrap([self.t1, self.t2])
        self.snapshot_before = self.group._snapshot()
        self.group.rotate(np.pi / 2., self.snapshot_before)
        self.cmd = RotateCommand(self.group, self.snapshot_before)
        self.stack = QUndoStack()
        self.stack.push(self.cmd)

    def test_first_redo_is_noop(self):
        self.assertAlmostEqual(self.t1._r[0], 0., places=5)
        self.assertAlmostEqual(self.t1._r[1], 4., places=5)

    def test_undo_restores_original_positions(self):
        self.stack.undo()
        self.assertAlmostEqual(self.t1._r[0], 4., places=5)
        self.assertAlmostEqual(self.t1._r[1], 0., places=5)

    def test_redo_after_undo_restores_rotated_positions(self):
        self.stack.undo()
        self.stack.redo()
        self.assertAlmostEqual(self.t1._r[0], 0., places=5)
        self.assertAlmostEqual(self.t1._r[1], 4., places=5)

    def test_text(self):
        self.assertEqual(self.cmd.text(), 'Rotate group')


class TestWheelCommand(unittest.TestCase):

    def setUp(self):
        self.trap = QTrap(r=(0., 0., 10.), phase=0.)
        self.trap._r[2] += 1.
        self.cmd = WheelCommand(self.trap, 1.)
        self.stack = QUndoStack()
        self.stack.push(self.cmd)

    def test_first_redo_is_noop(self):
        self.assertAlmostEqual(self.trap.z, 11.)

    def test_undo_reverses_scroll(self):
        self.stack.undo()
        self.assertAlmostEqual(self.trap.z, 10.)

    def test_redo_after_undo_reapplies(self):
        self.stack.undo()
        self.stack.redo()
        self.assertAlmostEqual(self.trap.z, 11.)

    def test_text(self):
        self.assertEqual(self.cmd.text(), 'Scroll trap z')


class TestWheelCommandMerge(unittest.TestCase):

    def setUp(self):
        self.trap = QTrap(r=(0., 0., 0.), phase=0.)
        self.stack = QUndoStack()
        # Three wheel events each incrementing z by 1
        for _ in range(3):
            self.trap._r[2] += 1.
            self.stack.push(WheelCommand(self.trap, 1.))

    def test_commands_merged_into_one(self):
        self.assertEqual(self.stack.count(), 1)

    def test_merged_dz(self):
        cmd = self.stack.command(0)
        self.assertAlmostEqual(cmd._dz, 3.)

    def test_z_after_scroll(self):
        self.assertAlmostEqual(self.trap.z, 3.)

    def test_undo_reverses_all_scrolls(self):
        self.stack.undo()
        self.assertAlmostEqual(self.trap.z, 0.)

    def test_redo_reapplies_all_scrolls(self):
        self.stack.undo()
        self.stack.redo()
        self.assertAlmostEqual(self.trap.z, 3.)


class TestWheelCommandNoMergeAcrossGroups(unittest.TestCase):

    def test_different_groups_not_merged(self):
        trap_a = QTrap(r=(0., 0., 0.), phase=0.)
        trap_b = QTrap(r=(1., 1., 0.), phase=0.)
        stack = QUndoStack()
        trap_a._r[2] += 1.
        stack.push(WheelCommand(trap_a, 1.))
        trap_b._r[2] += 1.
        stack.push(WheelCommand(trap_b, 1.))
        self.assertEqual(stack.count(), 2)


class TestLockCommand(unittest.TestCase):

    def setUp(self):
        self.overlay = make_overlay()
        self.trap = QTrap(r=(5., 5., 0.), phase=0.)
        self.overlay._addTrap(self.trap)
        self.stack = self.overlay._undoStack

    def test_redo_locks_trap(self):
        self.stack.push(LockCommand(self.overlay, self.trap))
        self.assertTrue(self.trap.locked)

    def test_undo_unlocks_trap(self):
        self.stack.push(LockCommand(self.overlay, self.trap))
        self.stack.undo()
        self.assertFalse(self.trap.locked)

    def test_redo_after_undo_locks_again(self):
        self.stack.push(LockCommand(self.overlay, self.trap))
        self.stack.undo()
        self.stack.redo()
        self.assertTrue(self.trap.locked)

    def test_redo_unlocks_already_locked(self):
        self.trap.locked = True
        self.stack.push(LockCommand(self.overlay, self.trap))
        self.assertFalse(self.trap.locked)

    def test_undo_relocks_after_unlock(self):
        self.trap.locked = True
        self.stack.push(LockCommand(self.overlay, self.trap))
        self.stack.undo()
        self.assertTrue(self.trap.locked)

    def test_text_lock(self):
        cmd = LockCommand(self.overlay, self.trap)
        self.assertEqual(cmd.text(), 'Lock trap')

    def test_text_unlock(self):
        self.trap.locked = True
        cmd = LockCommand(self.overlay, self.trap)
        self.assertEqual(cmd.text(), 'Unlock trap')

    def test_redo_sets_static_brush(self):
        self.stack.push(LockCommand(self.overlay, self.trap))
        spot = self.overlay.points()[0]
        self.assertEqual(spot.brush().color(),
                         self.overlay.brush[self.overlay.State.STATIC].color())

    def test_undo_restores_normal_brush(self):
        self.stack.push(LockCommand(self.overlay, self.trap))
        self.stack.undo()
        spot = self.overlay.points()[0]
        self.assertEqual(spot.brush().color(),
                         self.overlay.brush[self.overlay.State.NORMAL].color())


class TestUndoStackIntegration(unittest.TestCase):
    '''Integration: add, move, and remove via overlay and undo.'''

    def setUp(self):
        from pyqtgraph.Qt import QtCore
        self.overlay = make_overlay()
        self.stack = self.overlay._undoStack
        self.overlay.addTrap(QtCore.QPointF(5., 5.))

    def test_undo_add_empties_overlay(self):
        self.stack.undo()
        self.assertEqual(len(self.overlay._traps), 0)

    def test_redo_add_restores_overlay(self):
        self.stack.undo()
        self.stack.redo()
        self.assertEqual(len(self.overlay._traps), 1)

    def test_clear_traps_clears_stack(self):
        self.overlay.clearTraps()
        self.assertFalse(self.stack.canUndo())


if __name__ == '__main__':
    unittest.main()



'''Unit tests for MoveTraps.'''
import math
import unittest

import numpy as np
from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.MoveTraps import MoveTraps

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


# ------------------------------------------------------------------
# Test helpers

class FakeTrap:
    '''Minimal stand-in for QTrap: stores 3-D position as ndarray.'''

    def __init__(self, r=(0., 0., 0.)):
        self._r = np.array(r, dtype=float)

    @property
    def r(self):
        return self._r.copy()

    @r.setter
    def r(self, value):
        self._r[:] = value

    def leaves(self):
        yield self


class FakeOverlay:
    '''Minimal stand-in for QTrapOverlay: iterates over its traps.'''

    def __init__(self, *traps):
        self._traps = list(traps)
        self.marked = []   # tests may override to simulate marked traps

    def __iter__(self):
        return iter(self._traps)


# ------------------------------------------------------------------

class TestMoveTrapsInit(unittest.TestCase):

    def test_default_displacements_zero(self):
        task = MoveTraps()
        self.assertEqual(task.dx, 0.)
        self.assertEqual(task.dy, 0.)
        self.assertEqual(task.dz, 0.)

    def test_default_step(self):
        task = MoveTraps()
        self.assertEqual(task.step, 1.)

    def test_explicit_params_stored(self):
        task = MoveTraps(dx=10., dy=20., dz=5., step=2.)
        self.assertEqual(task.dx, 10.)
        self.assertEqual(task.dy, 20.)
        self.assertEqual(task.dz,  5.)
        self.assertEqual(task.step, 2.)

    def test_duration_keyword_raises(self):
        with self.assertRaises(TypeError):
            MoveTraps(dx=10., duration=5)

    def test_registered_in_registry(self):
        self.assertIn('MoveTraps', QTask._registry)

    def test_parameters_declared(self):
        names = [p['name'] for p in MoveTraps.parameters]
        self.assertIn('dx',   names)
        self.assertIn('dy',   names)
        self.assertIn('dz',   names)
        self.assertIn('step', names)

    def test_initial_state_pending(self):
        self.assertEqual(MoveTraps().state, QTask.State.PENDING)


class TestMoveTrapsDuration(unittest.TestCase):

    def test_zero_displacement_gives_duration_one(self):
        task = MoveTraps(dx=0., dy=0., dz=0.)
        self.assertEqual(task.duration, 1)

    def test_dx_only_duration(self):
        task = MoveTraps(dx=10., step=2.)
        self.assertEqual(task.duration, 5)

    def test_dy_only_duration(self):
        task = MoveTraps(dy=9., step=3.)
        self.assertEqual(task.duration, 3)

    def test_dz_only_duration(self):
        task = MoveTraps(dz=7., step=2.)
        self.assertEqual(task.duration, 4)   # ceil(7/2) = 4

    def test_3d_displacement_uses_l2_norm(self):
        dx, dy, dz, step = 3., 4., 0., 1.
        expected = math.ceil(math.sqrt(dx**2 + dy**2 + dz**2) / step)
        task = MoveTraps(dx=dx, dy=dy, dz=dz, step=step)
        self.assertEqual(task.duration, expected)

    def test_fractional_distance_rounds_up(self):
        task = MoveTraps(dx=5., step=3.)
        self.assertEqual(task.duration, 2)   # ceil(5/3) = 2

    def test_setting_dx_updates_duration(self):
        task = MoveTraps(dx=0., dy=0., dz=0., step=1.)
        task.dx = 10.
        self.assertEqual(task.duration, 10)

    def test_setting_dy_updates_duration(self):
        task = MoveTraps(step=2.)
        task.dy = 6.
        self.assertEqual(task.duration, 3)

    def test_setting_dz_updates_duration(self):
        task = MoveTraps(step=5.)
        task.dz = 10.
        self.assertEqual(task.duration, 2)

    def test_setting_step_updates_duration(self):
        task = MoveTraps(dx=10., step=1.)
        self.assertEqual(task.duration, 10)
        task.step = 5.
        self.assertEqual(task.duration, 2)

    def test_duration_minimum_is_one(self):
        task = MoveTraps(dx=0., step=1.)
        self.assertGreaterEqual(task.duration, 1)


class TestMoveTrapsExecution(unittest.TestCase):

    def setUp(self):
        self.trap = FakeTrap(r=(10., 20., 0.))
        self.overlay = FakeOverlay(self.trap)

    def _run(self, task):
        '''Start and step task to completion.'''
        task._start()
        for _ in range(task.duration):
            task._step()

    def test_initialize_records_start_positions(self):
        task = MoveTraps(overlay=self.overlay, dx=5.)
        task._start()
        task._step()          # calls initialize then process(0)
        self.assertIn(self.trap, task._starts)
        np.testing.assert_array_almost_equal(
            task._starts[self.trap], [10., 20., 0.])

    def test_final_position_equals_r0_plus_displacement(self):
        task = MoveTraps(overlay=self.overlay, dx=5., dy=-3., dz=2., step=1.)
        self._run(task)
        np.testing.assert_array_almost_equal(
            self.trap._r, [15., 17., 2.])

    def test_position_interpolates_linearly(self):
        task = MoveTraps(overlay=self.overlay, dx=10., step=5.)
        # duration = 2 frames; after frame 0 t=0.5, after frame 1 t=1.0
        task._start()
        task._step()          # initialize + process(0): t = 1/2
        np.testing.assert_array_almost_equal(
            self.trap._r, [15., 20., 0.])
        task._step()          # process(1): t = 2/2 = 1.0
        np.testing.assert_array_almost_equal(
            self.trap._r, [20., 20., 0.])

    def test_multiple_traps_all_moved(self):
        t1 = FakeTrap(r=(0., 0., 0.))
        t2 = FakeTrap(r=(10., 10., 0.))
        overlay = FakeOverlay(t1, t2)
        task = MoveTraps(overlay=overlay, dx=5., step=5.)
        self._run(task)
        np.testing.assert_array_almost_equal(t1._r, [5., 0., 0.])
        np.testing.assert_array_almost_equal(t2._r, [15., 10., 0.])

    def test_empty_overlay_does_not_raise(self):
        task = MoveTraps(overlay=FakeOverlay(), dx=5.)
        self._run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_completes_after_duration_frames(self):
        task = MoveTraps(overlay=self.overlay, dx=10., step=2.)
        self._run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_zero_displacement_completes_without_moving(self):
        task = MoveTraps(overlay=self.overlay, dx=0., dy=0., dz=0.)
        r_before = self.trap._r.copy()
        self._run(task)
        np.testing.assert_array_equal(self.trap._r, r_before)

    def test_finished_signal_emitted(self):
        task = MoveTraps(overlay=self.overlay, dx=5., step=5.)
        spy = QtTest.QSignalSpy(task.finished)
        self._run(task)
        self.assertEqual(len(spy), 1)


class TestMoveTrapsParamSync(unittest.TestCase):

    def setUp(self):
        from QHOT.lib.tasks.QTaskTree import QTaskTree
        self.QTaskTree = QTaskTree

    def test_editing_dx_updates_attribute_and_duration(self):
        task = MoveTraps(dx=0., step=1.)
        tree = self.QTaskTree(task)
        tree._params.child('dx').setValue(10.)
        self.assertAlmostEqual(task.dx, 10.)
        self.assertEqual(task.duration, 10)

    def test_editing_step_updates_duration(self):
        task = MoveTraps(dx=10., step=1.)
        tree = self.QTaskTree(task)
        tree._params.child('step').setValue(2.)
        self.assertAlmostEqual(task.step, 2.)
        self.assertEqual(task.duration, 5)


class TestMoveTrapsSelectedOnly(unittest.TestCase):

    def setUp(self):
        self.t1 = FakeTrap(r=(0., 0., 0.))
        self.t2 = FakeTrap(r=(10., 10., 0.))

    def _make_overlay_with_marked(self, *marked):
        '''Overlay that iterates both traps; marked returns only *marked.'''
        overlay = FakeOverlay(self.t1, self.t2)
        overlay.marked = iter(marked)
        return overlay

    def _run(self, task):
        task._start()
        for _ in range(task.duration):
            task._step()

    def test_selected_only_false_moves_all_traps(self):
        overlay = FakeOverlay(self.t1, self.t2)
        task = MoveTraps(overlay=overlay, dx=5., step=5.,
                         selected_only=False)
        self._run(task)
        np.testing.assert_array_almost_equal(self.t1._r, [5., 0., 0.])
        np.testing.assert_array_almost_equal(self.t2._r, [15., 10., 0.])

    def test_selected_only_true_moves_only_marked_traps(self):
        overlay = FakeOverlay(self.t1, self.t2)
        overlay.marked = [self.t1]     # only t1 is marked
        task = MoveTraps(overlay=overlay, dx=5., step=5.,
                         selected_only=True)
        self._run(task)
        np.testing.assert_array_almost_equal(self.t1._r, [5., 0., 0.])
        np.testing.assert_array_almost_equal(self.t2._r, [10., 10., 0.])

    def test_selected_only_true_empty_marked_moves_nothing(self):
        overlay = FakeOverlay(self.t1, self.t2)
        overlay.marked = []
        task = MoveTraps(overlay=overlay, dx=5., step=5.,
                         selected_only=True)
        self._run(task)
        np.testing.assert_array_equal(self.t1._r, [0., 0., 0.])
        np.testing.assert_array_equal(self.t2._r, [10., 10., 0.])

    def test_selected_only_default_is_false(self):
        task = MoveTraps()
        self.assertFalse(task.selected_only)

    def test_selected_only_in_parameters(self):
        names = [p['name'] for p in MoveTraps.parameters]
        self.assertIn('selected_only', names)

    def test_selected_only_round_trip(self):
        task = MoveTraps(dx=5., selected_only=True)
        restored = QTask.from_dict(task.to_dict())
        self.assertTrue(restored.selected_only)


class TestMoveTrapsSerialization(unittest.TestCase):

    def test_to_dict_includes_type(self):
        d = MoveTraps(dx=5., dy=3., dz=1., step=2.).to_dict()
        self.assertEqual(d['type'], 'MoveTraps')

    def test_to_dict_includes_all_params(self):
        d = MoveTraps(dx=5., dy=3., dz=1., step=2.).to_dict()
        self.assertAlmostEqual(d['dx'],   5.)
        self.assertAlmostEqual(d['dy'],   3.)
        self.assertAlmostEqual(d['dz'],   1.)
        self.assertAlmostEqual(d['step'], 2.)

    def test_round_trip(self):
        task = MoveTraps(dx=5., dy=3., dz=1., step=2.)
        d = task.to_dict()
        restored = QTask.from_dict(d)
        self.assertIsInstance(restored, MoveTraps)
        self.assertAlmostEqual(restored.dx,   5.)
        self.assertAlmostEqual(restored.dy,   3.)
        self.assertAlmostEqual(restored.dz,   1.)
        self.assertAlmostEqual(restored.step, 2.)

    def test_round_trip_preserves_duration(self):
        task = MoveTraps(dx=5., dy=3., dz=1., step=2.)
        restored = QTask.from_dict(task.to_dict())
        self.assertEqual(restored.duration, task.duration)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for StopRecording.'''
import unittest

from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.StopRecording import StopRecording

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class TestStopRecordingInit(unittest.TestCase):

    def setUp(self):
        from unittest.mock import MagicMock
        self.dvr = MagicMock()

    def test_initial_state_is_pending(self):
        task = StopRecording(dvr=self.dvr)
        self.assertEqual(task.state, QTask.State.PENDING)

    def test_duration_is_zero(self):
        task = StopRecording(dvr=self.dvr)
        self.assertEqual(task.duration, 0)

    def test_no_parameters(self):
        self.assertEqual(StopRecording.parameters, [])

    def test_accepts_delay_kwarg(self):
        task = StopRecording(dvr=self.dvr, delay=3)
        self.assertEqual(task.delay, 3)


class TestStopRecordingExecution(unittest.TestCase):

    def setUp(self):
        from unittest.mock import MagicMock
        self.dvr = MagicMock()

    def test_stop_button_clicked_on_initialize(self):
        task = StopRecording(dvr=self.dvr)
        task._start()
        task._step()
        self.dvr.stopButton.animateClick.assert_called_once()

    def test_completes_after_one_step(self):
        task = StopRecording(dvr=self.dvr)
        task._start()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_finished_emitted_after_one_step(self):
        task = StopRecording(dvr=self.dvr)
        spy = QtTest.QSignalSpy(task.finished)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)

    def test_stop_not_called_before_start(self):
        task = StopRecording(dvr=self.dvr)
        task._step()
        self.dvr.stopButton.animateClick.assert_not_called()

    def test_animateclick_exception_fails_task(self):
        self.dvr.stopButton.animateClick.side_effect = RuntimeError('no dvr')
        task = StopRecording(dvr=self.dvr)
        spy = QtTest.QSignalSpy(task.failed)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)
        self.assertEqual(task.state, QTask.State.FAILED)


class TestStopRecordingSerialization(unittest.TestCase):

    def setUp(self):
        from unittest.mock import MagicMock
        self.dvr = MagicMock()

    def test_to_dict_type(self):
        task = StopRecording(dvr=self.dvr)
        self.assertEqual(task.to_dict()['type'], 'StopRecording')

    def test_to_dict_delay(self):
        task = StopRecording(dvr=self.dvr, delay=2)
        self.assertEqual(task.to_dict()['delay'], 2)

    def test_round_trip(self):
        task = StopRecording(dvr=self.dvr)
        d = task.to_dict()
        task2 = QTask.from_dict({**d, 'dvr': self.dvr})
        self.assertIsInstance(task2, StopRecording)
        self.assertEqual(task2.duration, 0)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for AddTweezer.'''
import unittest
from unittest.mock import MagicMock, call

from pyqtgraph.Qt import QtCore, QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.AddTweezer import AddTweezer

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class TestAddTweezerInit(unittest.TestCase):

    def setUp(self):
        self.overlay = MagicMock()
        self.cgh = MagicMock()
        self.cgh.xc = 320.
        self.cgh.yc = 240.

    def test_duration_is_zero(self):
        task = AddTweezer(overlay=self.overlay)
        self.assertEqual(task.duration, 0)

    def test_initial_state_is_pending(self):
        task = AddTweezer(overlay=self.overlay)
        self.assertEqual(task.state, QTask.State.PENDING)

    def test_explicit_x_y_stored(self):
        task = AddTweezer(overlay=self.overlay, x=100., y=200.)
        self.assertEqual(task.x, 100.)
        self.assertEqual(task.y, 200.)

    def test_default_position_uses_cgh_center(self):
        task = AddTweezer(overlay=self.overlay, cgh=self.cgh)
        self.assertEqual(task.x, 320.)
        self.assertEqual(task.y, 240.)

    def test_default_position_without_cgh_is_origin(self):
        task = AddTweezer(overlay=self.overlay)
        self.assertEqual(task.x, 0.)
        self.assertEqual(task.y, 0.)

    def test_registered_in_registry(self):
        self.assertIn('AddTweezer', QTask._registry)

    def test_parameters_declared(self):
        names = [p['name'] for p in AddTweezer.parameters]
        self.assertIn('x', names)
        self.assertIn('y', names)

    def test_position_attributes_set(self):
        task = AddTweezer(overlay=self.overlay, x=50., y=75.)
        self.assertAlmostEqual(task.x, 50.)
        self.assertAlmostEqual(task.y, 75.)


class TestAddTweezerExecution(unittest.TestCase):

    def setUp(self):
        self.overlay = MagicMock()
        self.cgh = MagicMock()
        self.cgh.xc = 320.
        self.cgh.yc = 240.

    def test_add_trap_called_on_step(self):
        task = AddTweezer(overlay=self.overlay, x=10., y=20.)
        task._start()
        task._step()
        self.overlay.addTrap.assert_called_once()

    def test_add_trap_called_with_correct_position(self):
        task = AddTweezer(overlay=self.overlay, x=10., y=20.)
        task._start()
        task._step()
        args = self.overlay.addTrap.call_args[0]
        pos = args[0]
        self.assertIsInstance(pos, QtCore.QPointF)
        self.assertAlmostEqual(pos.x(), 10.)
        self.assertAlmostEqual(pos.y(), 20.)

    def test_completes_after_one_step(self):
        task = AddTweezer(overlay=self.overlay, x=0., y=0.)
        task._start()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_add_trap_not_called_before_start(self):
        task = AddTweezer(overlay=self.overlay, x=0., y=0.)
        task._step()
        self.overlay.addTrap.assert_not_called()

    def test_add_trap_called_exactly_once(self):
        task = AddTweezer(overlay=self.overlay, x=0., y=0.)
        task._start()
        task._step()
        task._step()
        self.overlay.addTrap.assert_called_once()

    def test_process_never_called(self):
        task = AddTweezer(overlay=self.overlay, x=0., y=0.)
        task.process = MagicMock()
        task._start()
        task._step()
        task.process.assert_not_called()

    def test_finished_signal_emitted(self):
        task = AddTweezer(overlay=self.overlay, x=0., y=0.)
        spy = QtTest.QSignalSpy(task.finished)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)

    def test_overlay_exception_fails_task(self):
        self.overlay.addTrap.side_effect = RuntimeError('scene error')
        task = AddTweezer(overlay=self.overlay, x=0., y=0.)
        spy = QtTest.QSignalSpy(task.failed)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)
        self.assertEqual(task.state, QTask.State.FAILED)


class TestAddTweezerParamSync(unittest.TestCase):

    def setUp(self):
        from QHOT.lib.tasks.QTaskTree import QTaskTree
        self.QTaskTree = QTaskTree

    def test_editing_x_param_updates_attribute(self):
        task = AddTweezer(x=10., y=20.)
        tree = self.QTaskTree(task)
        tree._params.child('x').setValue(99.)
        self.assertAlmostEqual(task.x, 99.)

    def test_editing_y_param_updates_attribute(self):
        task = AddTweezer(x=10., y=20.)
        tree = self.QTaskTree(task)
        tree._params.child('y').setValue(55.)
        self.assertAlmostEqual(task.y, 55.)

    def test_initialize_uses_updated_param_value(self):
        overlay = MagicMock()
        task = AddTweezer(overlay=overlay, x=10., y=20.)
        tree = self.QTaskTree(task)
        tree._params.child('x').setValue(50.)
        tree._params.child('y').setValue(75.)
        task._start()
        task._step()
        args = overlay.addTrap.call_args[0]
        self.assertAlmostEqual(args[0].x(), 50.)
        self.assertAlmostEqual(args[0].y(), 75.)


class TestAddTweezerSerialization(unittest.TestCase):

    def test_to_dict_includes_type(self):
        task = AddTweezer(x=10., y=20.)
        d = task.to_dict()
        self.assertEqual(d['type'], 'AddTweezer')

    def test_to_dict_includes_position(self):
        task = AddTweezer(x=10., y=20.)
        d = task.to_dict()
        self.assertAlmostEqual(d['x'], 10.)
        self.assertAlmostEqual(d['y'], 20.)

    def test_round_trip(self):
        task = AddTweezer(x=100., y=150.)
        d = task.to_dict()
        restored = QTask.from_dict(d)
        self.assertIsInstance(restored, AddTweezer)
        self.assertAlmostEqual(restored.x, 100.)
        self.assertAlmostEqual(restored.y, 150.)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for Snapshot.'''
import unittest
from unittest.mock import MagicMock, call

from pyqtgraph.Qt import QtWidgets

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.Snapshot import Snapshot

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


def _run(task):
    task._start()
    task._step()


class TestSnapshotInit(unittest.TestCase):

    def test_registered_in_registry(self):
        self.assertIn('Snapshot', QTask._registry)

    def test_duration_is_zero(self):
        self.assertEqual(Snapshot(image=MagicMock()).duration, 0)

    def test_default_filename_empty(self):
        self.assertEqual(Snapshot(image=MagicMock()).filename, '')

    def test_default_prefix(self):
        self.assertEqual(Snapshot(image=MagicMock()).prefix, 'snapshot')

    def test_explicit_filename_stored(self):
        task = Snapshot(image=MagicMock(), filename='/tmp/snap.png')
        self.assertEqual(task.filename, '/tmp/snap.png')

    def test_explicit_prefix_stored(self):
        task = Snapshot(image=MagicMock(), prefix='frame')
        self.assertEqual(task.prefix, 'frame')

    def test_parameters_declared(self):
        names = [p['name'] for p in Snapshot.parameters]
        self.assertIn('filename', names)
        self.assertIn('prefix',   names)


class TestSnapshotExecution(unittest.TestCase):

    def setUp(self):
        self.image = MagicMock()
        self.save  = MagicMock()

    def test_save_image_called_with_image_item(self):
        task = Snapshot(image=self.image, save=self.save)
        _run(task)
        self.save.image.assert_called_once()
        args, kwargs = self.save.image.call_args
        self.assertIs(args[0], self.image)

    def test_explicit_filename_passed_to_save(self):
        task = Snapshot(image=self.image, save=self.save,
                        filename='/tmp/snap.png')
        _run(task)
        _, kwargs = self.save.image.call_args
        self.assertEqual(kwargs['filename'], '/tmp/snap.png')

    def test_empty_filename_passes_none(self):
        task = Snapshot(image=self.image, save=self.save, filename='')
        _run(task)
        _, kwargs = self.save.image.call_args
        self.assertIsNone(kwargs['filename'])

    def test_prefix_passed_to_save(self):
        task = Snapshot(image=self.image, save=self.save, prefix='frame')
        _run(task)
        _, kwargs = self.save.image.call_args
        self.assertEqual(kwargs['prefix'], 'frame')

    def test_completes_after_one_step(self):
        task = Snapshot(image=self.image, save=self.save)
        _run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



'''Unit tests for StartRecording.'''
import unittest
from unittest.mock import MagicMock, call

from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.StartRecording import StartRecording

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


def _make_dvr(filename='', nframes=10_000, nskip=1):
    '''Return a mock DVR whose spinboxes report sensible defaults.'''
    dvr = MagicMock()
    dvr.filename = filename
    dvr.nframes.value.return_value = nframes
    dvr.nskip.value.return_value = nskip
    return dvr


class TestStartRecordingInit(unittest.TestCase):

    def test_initial_state_is_pending(self):
        task = StartRecording(dvr=_make_dvr())
        self.assertEqual(task.state, QTask.State.PENDING)

    def test_duration_is_zero(self):
        task = StartRecording(dvr=_make_dvr())
        self.assertEqual(task.duration, 0)

    def test_nframes_forwarded_to_dvr(self):
        dvr = _make_dvr()
        StartRecording(dvr=dvr, nframes=500)
        dvr.nframes.setValue.assert_called_with(500)

    def test_nskip_forwarded_to_dvr(self):
        dvr = _make_dvr()
        StartRecording(dvr=dvr, nskip=3)
        dvr.nskip.setValue.assert_called_with(3)

    def test_filename_forwarded_to_dvr(self):
        dvr = _make_dvr()
        StartRecording(dvr=dvr, filename='clip.mkv')
        self.assertEqual(dvr.filename, 'clip.mkv')

    def test_empty_filename_leaves_dvr_unchanged(self):
        dvr = _make_dvr(filename='existing.mkv')
        StartRecording(dvr=dvr, filename='')
        self.assertEqual(dvr.filename, 'existing.mkv')


class TestStartRecordingProperties(unittest.TestCase):

    def setUp(self):
        self.dvr = _make_dvr(filename='out.mkv', nframes=200, nskip=2)
        self.task = StartRecording(dvr=self.dvr)

    def test_filename_reads_dvr(self):
        self.assertEqual(self.task.filename, 'out.mkv')

    def test_filename_setter_updates_dvr(self):
        self.task.filename = 'new.mkv'
        self.assertEqual(self.dvr.filename, 'new.mkv')

    def test_filename_setter_ignores_empty(self):
        self.task.filename = ''
        self.assertEqual(self.dvr.filename, 'out.mkv')

    def test_nframes_reads_dvr_spinbox(self):
        self.assertEqual(self.task.nframes, 200)

    def test_nframes_setter_updates_dvr_spinbox(self):
        self.task.nframes = 999
        self.dvr.nframes.setValue.assert_called_with(999)

    def test_nframes_setter_coerces_to_int(self):
        self.task.nframes = 42.9
        self.dvr.nframes.setValue.assert_called_with(42)

    def test_nskip_reads_dvr_spinbox(self):
        self.assertEqual(self.task.nskip, 2)

    def test_nskip_setter_updates_dvr_spinbox(self):
        self.task.nskip = 4
        self.dvr.nskip.setValue.assert_called_with(4)

    def test_nskip_setter_coerces_to_int(self):
        self.task.nskip = 2.7
        self.dvr.nskip.setValue.assert_called_with(2)


class TestStartRecordingExecution(unittest.TestCase):

    def setUp(self):
        self.dvr = _make_dvr()

    def test_record_button_clicked_on_initialize(self):
        task = StartRecording(dvr=self.dvr)
        task._start()
        task._step()
        self.dvr.recordButton.animateClick.assert_called_once()

    def test_completes_after_one_step(self):
        task = StartRecording(dvr=self.dvr)
        task._start()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_finished_emitted_after_one_step(self):
        task = StartRecording(dvr=self.dvr)
        spy = QtTest.QSignalSpy(task.finished)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)

    def test_record_not_called_before_start(self):
        task = StartRecording(dvr=self.dvr)
        task._step()
        self.dvr.recordButton.animateClick.assert_not_called()

    def test_animateclick_exception_fails_task(self):
        self.dvr.recordButton.animateClick.side_effect = RuntimeError('no dvr')
        task = StartRecording(dvr=self.dvr)
        spy = QtTest.QSignalSpy(task.failed)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)
        self.assertEqual(task.state, QTask.State.FAILED)


class TestStartRecordingSerialization(unittest.TestCase):

    def setUp(self):
        self.dvr = _make_dvr()

    def test_to_dict_type(self):
        task = StartRecording(dvr=self.dvr)
        self.assertEqual(task.to_dict()['type'], 'StartRecording')

    def test_to_dict_filename(self):
        dvr = _make_dvr(filename='out.mkv')
        task = StartRecording(dvr=dvr, filename='out.mkv')
        self.assertEqual(task.to_dict()['filename'], 'out.mkv')

    def test_to_dict_nframes(self):
        dvr = _make_dvr(nframes=300)
        task = StartRecording(dvr=dvr, nframes=300)
        self.assertEqual(task.to_dict()['nframes'], 300)

    def test_to_dict_nskip(self):
        dvr = _make_dvr(nskip=2)
        task = StartRecording(dvr=dvr, nskip=2)
        self.assertEqual(task.to_dict()['nskip'], 2)

    def test_round_trip(self):
        dvr = _make_dvr(filename='clip.mkv', nframes=100, nskip=2)
        task = StartRecording(dvr=dvr, filename='clip.mkv',
                              nframes=100, nskip=2)
        d = task.to_dict()
        task2 = QTask.from_dict({**d, 'dvr': dvr})
        self.assertIsInstance(task2, StartRecording)
        dvr.nframes.setValue.assert_called_with(100)
        dvr.nskip.setValue.assert_called_with(2)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for ClearTraps.'''
import unittest
from unittest.mock import MagicMock

from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.ClearTraps import ClearTraps

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class TestClearTrapsInit(unittest.TestCase):

    def setUp(self):
        self.overlay = MagicMock()

    def test_duration_is_zero(self):
        task = ClearTraps(self.overlay)
        self.assertEqual(task.duration, 0)

    def test_duration_cannot_be_overridden(self):
        # duration=0 is set unconditionally; any caller-supplied
        # duration is silently overridden by the super().__init__ call
        task = ClearTraps(self.overlay)
        self.assertEqual(task.duration, 0)

    def test_initial_state_is_pending(self):
        task = ClearTraps(self.overlay)
        self.assertEqual(task.state, QTask.State.PENDING)


class TestClearTrapsExecution(unittest.TestCase):

    def setUp(self):
        self.overlay = MagicMock()

    def test_clear_traps_called_on_step(self):
        task = ClearTraps(self.overlay)
        task._start()
        task._step()
        self.overlay.clearTraps.assert_called_once()

    def test_completes_after_one_step(self):
        task = ClearTraps(self.overlay)
        task._start()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_process_never_called(self):
        task = ClearTraps(self.overlay)
        task.process = MagicMock()
        task._start()
        task._step()
        task.process.assert_not_called()

    def test_clear_traps_not_called_before_start(self):
        task = ClearTraps(self.overlay)
        task._step()
        self.overlay.clearTraps.assert_not_called()

    def test_clear_traps_called_exactly_once(self):
        task = ClearTraps(self.overlay)
        task._start()
        task._step()
        task._step()
        task._step()
        self.overlay.clearTraps.assert_called_once()

    def test_finished_signal_emitted(self):
        task = ClearTraps(self.overlay)
        spy = QtTest.QSignalSpy(task.finished)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)

    def test_overlay_exception_fails_task(self):
        self.overlay.clearTraps.side_effect = RuntimeError('no traps')
        task = ClearTraps(self.overlay)
        spy = QtTest.QSignalSpy(task.failed)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)
        self.assertEqual(task.state, QTask.State.FAILED)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for Record.'''
import unittest
from unittest.mock import MagicMock, PropertyMock

from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.Record import Record

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class TestRecordInit(unittest.TestCase):

    def setUp(self):
        self.dvr = MagicMock()

    def test_initial_state_is_pending(self):
        task = Record(dvr=self.dvr)
        self.assertEqual(task.state, QTask.State.PENDING)

    def test_filename_default_empty(self):
        task = Record(dvr=self.dvr)
        self.assertEqual(task.filename, '')

    def test_filename_stored(self):
        task = Record(dvr=self.dvr, filename='out.mkv')
        self.assertEqual(task.filename, 'out.mkv')

    def test_nframes_forwarded(self):
        task = Record(dvr=self.dvr, nframes=120)
        self.assertEqual(task.duration, 120)

    def test_nframes_default_unlimited(self):
        task = Record(dvr=self.dvr)
        self.assertIsNone(task.duration)


class TestRecordExecution(unittest.TestCase):

    def setUp(self):
        self.dvr = MagicMock()

    def test_record_called_on_initialize(self):
        task = Record(dvr=self.dvr, nframes=1)
        task._start()
        task._step()
        self.dvr.record.assert_called_once()

    def test_stop_called_on_complete(self):
        task = Record(dvr=self.dvr, nframes=1)
        task._start()
        task._step()
        self.dvr.stop.assert_called_once()

    def test_record_not_called_before_start(self):
        task = Record(dvr=self.dvr)
        task._step()
        self.dvr.record.assert_not_called()

    def test_filename_set_before_record(self):
        calls = []
        self.dvr.record.side_effect = lambda: calls.append('record')
        type(self.dvr).filename = PropertyMock(
            side_effect=lambda v: calls.append(f'filename={v}'))
        task = Record(dvr=self.dvr, nframes=1, filename='test.mkv')
        task._start()
        task._step()
        filename_idx = next(
            (i for i, c in enumerate(calls) if c.startswith('filename')),
            None)
        record_idx = next(
            (i for i, c in enumerate(calls) if c == 'record'), None)
        if filename_idx is not None and record_idx is not None:
            self.assertLess(filename_idx, record_idx)

    def test_filename_not_set_when_empty(self):
        task = Record(dvr=self.dvr, nframes=1)
        task._start()
        task._step()
        # dvr.filename should never have been assigned
        assigned = any(
            str(c).startswith('call.filename =')
            for c in self.dvr.mock_calls)
        self.assertFalse(assigned)

    def test_stop_not_called_mid_recording(self):
        task = Record(dvr=self.dvr, nframes=5)
        task._start()
        task._step()   # initialize → record
        task._step()   # process(1)
        self.dvr.stop.assert_not_called()

    def test_stop_called_after_duration_expires(self):
        task = Record(dvr=self.dvr, nframes=3)
        task._start()
        for _ in range(3):
            task._step()
        self.dvr.stop.assert_called_once()

    def test_stop_called_on_finish(self):
        task = Record(dvr=self.dvr)
        task._start()
        task._step()   # initialize → record
        task.finish()
        self.dvr.stop.assert_called_once()

    def test_record_exception_fails_task(self):
        self.dvr.record.side_effect = RuntimeError('device busy')
        task = Record(dvr=self.dvr)
        spy = QtTest.QSignalSpy(task.failed)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)
        self.assertEqual(task.state, QTask.State.FAILED)

    def test_stop_exception_fails_task(self):
        self.dvr.stop.side_effect = RuntimeError('stop failed')
        task = Record(dvr=self.dvr, nframes=1)
        spy = QtTest.QSignalSpy(task.failed)
        task._start()
        task._step()
        self.assertEqual(len(spy), 1)
        self.assertEqual(task.state, QTask.State.FAILED)

    def test_completes_after_duration(self):
        task = Record(dvr=self.dvr, nframes=2)
        task._start()
        task._step()
        task._step()
        self.assertEqual(task.state, QTask.State.COMPLETED)


class TestRecordSerialization(unittest.TestCase):

    def setUp(self):
        self.dvr = MagicMock()

    def test_to_dict_type(self):
        task = Record(dvr=self.dvr)
        self.assertEqual(task.to_dict()['type'], 'Record')

    def test_to_dict_filename(self):
        task = Record(dvr=self.dvr, filename='out.mkv')
        self.assertEqual(task.to_dict()['filename'], 'out.mkv')

    def test_to_dict_nframes(self):
        task = Record(dvr=self.dvr, nframes=60)
        self.assertEqual(task.to_dict()['nframes'], 60)

    def test_to_dict_delay(self):
        task = Record(dvr=self.dvr, delay=5)
        self.assertEqual(task.to_dict()['delay'], 5)

    def test_round_trip(self):
        task = Record(dvr=self.dvr, filename='clip.mkv', nframes=100)
        d = task.to_dict()
        task2 = QTask.from_dict({**d, 'dvr': self.dvr})
        self.assertEqual(task2.filename, 'clip.mkv')
        self.assertEqual(task2.nframes, 100)
        self.assertEqual(task2.duration, 100)


if __name__ == '__main__':
    unittest.main()



'''Unit tests for Move.'''
import math
import unittest

import numpy as np
from pyqtgraph.Qt import QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.Move import Move

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


# ------------------------------------------------------------------
# Helpers (shared with test_movetraps.py pattern)

class FakeTrap:
    def __init__(self, r=(0., 0., 0.)):
        self._r = np.array(r, dtype=float)

    @property
    def r(self):
        return self._r.copy()

    @r.setter
    def r(self, value):
        self._r[:] = value


class FakeOverlay:
    def __init__(self, *traps):
        self._traps = list(traps)
        self.marked = list(traps)   # default: all traps marked

    def __iter__(self):
        return iter(self._traps)


def _run(task, max_steps=10_000):
    '''Step task to completion; duration may be set inside initialize().'''
    task._start()
    for _ in range(max_steps):
        task._step()
        if task.state != QTask.State.RUNNING:
            break


# ------------------------------------------------------------------
# Init

class TestMoveInit(unittest.TestCase):

    def test_default_target_zero(self):
        task = Move()
        self.assertEqual(task.x, 0.)
        self.assertEqual(task.y, 0.)
        self.assertEqual(task.z, 0.)

    def test_default_step(self):
        self.assertEqual(Move().step, 1.)

    def test_explicit_params_stored(self):
        task = Move(x=10., y=20., z=5., step=2.)
        self.assertEqual(task.x, 10.)
        self.assertEqual(task.y, 20.)
        self.assertEqual(task.z,  5.)
        self.assertEqual(task.step, 2.)

    def test_duration_keyword_raises(self):
        with self.assertRaises(TypeError):
            Move(x=10., duration=5)

    def test_registered_in_registry(self):
        self.assertIn('Move', QTask._registry)

    def test_parameters_declared(self):
        names = [p['name'] for p in Move.parameters]
        for name in ('x', 'y', 'z', 'step'):
            self.assertIn(name, names)

    def test_initial_state_pending(self):
        self.assertEqual(Move().state, QTask.State.PENDING)


# ------------------------------------------------------------------
# Execution

class TestMoveExecution(unittest.TestCase):

    def setUp(self):
        self.trap = FakeTrap(r=(10., 20., 0.))
        self.overlay = FakeOverlay(self.trap)

    def test_trap_reaches_target(self):
        task = Move(overlay=self.overlay, x=50., y=80., z=0., step=1.)
        _run(task)
        np.testing.assert_array_almost_equal(self.trap._r, [50., 80., 0.])

    def test_3d_target(self):
        task = Move(overlay=self.overlay, x=10., y=20., z=5., step=1.)
        _run(task)
        np.testing.assert_array_almost_equal(self.trap._r, [10., 20., 5.])

    def test_multiple_traps_maintain_relative_config(self):
        t1 = FakeTrap(r=(0.,  0., 0.))
        t2 = FakeTrap(r=(10., 0., 0.))
        overlay = FakeOverlay(t1, t2)
        # centroid = (5, 0, 0); target = (15, 0, 0); shift = (+10, 0, 0)
        task = Move(overlay=overlay, x=15., y=0., z=0., step=1.)
        _run(task)
        np.testing.assert_array_almost_equal(t1._r, [10.,  0., 0.])
        np.testing.assert_array_almost_equal(t2._r, [20.,  0., 0.])

    def test_no_marked_traps_completes_immediately(self):
        overlay = FakeOverlay(self.trap)
        overlay.marked = []
        task = Move(overlay=overlay, x=50., y=50.)
        task._start()
        task._step()   # initialize sets duration=0; _step completes it
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_no_marked_traps_does_not_move(self):
        overlay = FakeOverlay(self.trap)
        overlay.marked = []
        r_before = self.trap._r.copy()
        task = Move(overlay=overlay, x=50., y=50.)
        task._start()
        task._step()
        np.testing.assert_array_equal(self.trap._r, r_before)

    def test_target_already_at_current_position_completes(self):
        task = Move(overlay=self.overlay, x=10., y=20., z=0.)
        _run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_finished_signal_emitted(self):
        task = Move(overlay=self.overlay, x=20., y=20., step=10.)
        spy = QtTest.QSignalSpy(task.finished)
        _run(task)
        self.assertEqual(len(spy), 1)

    def test_duration_scales_with_displacement(self):
        trap = FakeTrap(r=(10., 20., 0.))
        overlay = FakeOverlay(trap)
        task = Move(overlay=overlay, x=10., y=30., z=0., step=2.)
        # centroid at (10, 20, 0), target at (10, 30, 0), displacement = 10
        task._start()
        task._step()   # calls initialize, which sets duration
        self.assertEqual(task.duration, math.ceil(10. / 2.))


# ------------------------------------------------------------------
# Serialization

class TestMoveSerialization(unittest.TestCase):

    def test_to_dict_includes_type(self):
        self.assertEqual(Move(x=5.).to_dict()['type'], 'Move')

    def test_round_trip(self):
        task = Move(x=5., y=3., z=1., step=2.)
        restored = QTask.from_dict(task.to_dict())
        self.assertIsInstance(restored, Move)
        self.assertAlmostEqual(restored.x,    5.)
        self.assertAlmostEqual(restored.y,    3.)
        self.assertAlmostEqual(restored.z,    1.)
        self.assertAlmostEqual(restored.step, 2.)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



'''Unit tests for QTaskManager.'''
import unittest
from unittest.mock import MagicMock

from pyqtgraph.Qt import QtCore, QtWidgets, QtTest

from QHOT.lib.tasks.QTask import QTask
from QHOT.lib.tasks.QTaskManager import QTaskManager

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class MockScreen(QtCore.QObject):
    '''Minimal screen substitute with a rendered signal.'''
    rendered = QtCore.pyqtSignal()


class TestQTaskManagerInit(unittest.TestCase):

    def setUp(self):
        self.screen = MockScreen()

    def test_no_active_task_initially(self):
        mgr = QTaskManager(self.screen)
        self.assertIsNone(mgr.active)

    def test_queue_size_zero_initially(self):
        mgr = QTaskManager(self.screen)
        self.assertEqual(mgr.queue_size, 0)

    def test_background_empty_initially(self):
        mgr = QTaskManager(self.screen)
        self.assertEqual(mgr.background, [])

    def test_not_paused_initially(self):
        mgr = QTaskManager(self.screen)
        self.assertFalse(mgr.paused)

    def test_overlay_stored(self):
        overlay = MagicMock()
        mgr = QTaskManager(self.screen, overlay=overlay)
        self.assertIs(mgr.overlay, overlay)

    def test_cgh_stored(self):
        cgh = MagicMock()
        mgr = QTaskManager(self.screen, cgh=cgh)
        self.assertIs(mgr.cgh, cgh)


class TestQTaskManagerRegistration(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def test_first_blocking_task_becomes_active(self):
        task = QTask()
        self.manager.register(task)
        self.assertIs(self.manager.active_raw, task)

    def test_first_blocking_task_starts_running(self):
        task = QTask()
        self.manager.register(task)
        self.assertEqual(task.state, QTask.State.RUNNING)

    def test_second_blocking_task_queued(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.assertEqual(self.manager.queue_size, 1)

    def test_third_blocking_task_increments_queue(self):
        t1, t2, t3 = QTask(), QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.manager.register(t3)
        self.assertEqual(self.manager.queue_size, 2)

    def test_nonblocking_task_not_queued(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        self.assertEqual(self.manager.queue_size, 0)
        self.assertIsNone(self.manager.active)

    def test_nonblocking_task_starts_immediately(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        self.assertEqual(task.state, QTask.State.RUNNING)

    def test_nonblocking_task_in_background_list(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        self.assertIn(task, self.manager.background)

    def test_register_returns_task(self):
        task = QTask()
        result = self.manager.register(task)
        self.assertIs(result, task)

    def test_background_property_is_copy(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        bg = self.manager.background
        bg.clear()
        self.assertEqual(len(self.manager.background), 1)

    def test_queued_empty_initially(self):
        self.assertEqual(self.manager.queued, [])

    def test_queued_contains_pending_tasks(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        # t1 activated but not yet stepped — still in queued
        self.assertIn(t1, self.manager.queued)
        self.assertIn(t2, self.manager.queued)

    def test_queued_property_is_copy(self):
        t1, t2 = QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        q = self.manager.queued
        q.clear()
        self.assertEqual(len(self.manager.queued), 2)


class TestQTaskManagerFrameDispatch(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_frame_steps_active_blocking_task(self):
        task = QTask()
        task._step = MagicMock()
        self.manager.register(task)
        self._emit()
        task._step.assert_called_once()

    def test_frame_steps_background_task(self):
        task = QTask()
        task._step = MagicMock()
        self.manager.register(task, blocking=False)
        self._emit()
        task._step.assert_called_once()

    def test_pause_prevents_blocking_step(self):
        task = QTask()
        task._step = MagicMock()
        self.manager.register(task)
        self.manager.pause(True)
        self._emit()
        task._step.assert_not_called()

    def test_pause_prevents_background_step(self):
        task = QTask()
        task._step = MagicMock()
        self.manager.register(task, blocking=False)
        self.manager.pause(True)
        self._emit()
        task._step.assert_not_called()

    def test_resume_restores_dispatch(self):
        task = QTask()
        task._step = MagicMock()
        self.manager.register(task)
        self.manager.pause(True)
        self.manager.pause(False)
        self._emit()
        task._step.assert_called_once()

    def test_paused_property_reflects_state(self):
        self.manager.pause(True)
        self.assertTrue(self.manager.paused)
        self.manager.pause(False)
        self.assertFalse(self.manager.paused)

    def test_no_step_when_no_active_task(self):
        # Should not raise even with nothing registered
        self._emit()


class TestQTaskManagerSequencing(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_second_task_activates_after_first_completes(self):
        t1 = QTask(duration=1)
        t2 = QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit(1)
        self.assertEqual(t1.state, QTask.State.COMPLETED)
        self.assertIs(self.manager.active_raw, t2)

    def test_previous_passed_from_first_to_second(self):
        t1 = QTask(duration=1)
        t2 = QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self._emit(1)
        self.assertIs(t2.previous, t1)

    def test_previous_is_none_for_first_task(self):
        task = QTask()
        self.manager.register(task)
        self.assertIsNone(task.previous)

    def test_three_tasks_run_in_order(self):
        order = []
        tasks = [QTask(duration=1) for _ in range(3)]
        for i, t in enumerate(tasks):
            t.started.connect(lambda _i=i: order.append(_i))
            self.manager.register(t)
        self._emit(3)
        self.assertEqual(order, [0, 1, 2])

    def test_queue_depletes_as_tasks_complete(self):
        tasks = [QTask(duration=1) for _ in range(3)]
        for t in tasks:
            self.manager.register(t)
        self.assertEqual(self.manager.queue_size, 2)
        self._emit(1)
        self.assertEqual(self.manager.queue_size, 1)
        self._emit(1)
        self.assertEqual(self.manager.queue_size, 0)

    def test_active_is_none_after_all_complete(self):
        task = QTask(duration=1)
        self.manager.register(task)
        self._emit(1)
        self.assertIsNone(self.manager.active)

    def test_blocking_failure_clears_queue(self):
        t1, t2, t3 = QTask(), QTask(), QTask()
        self.manager.register(t1)
        self.manager.register(t2)
        self.manager.register(t3)
        t1.abort('test failure')
        self.assertEqual(self.manager.queue_size, 0)
        self.assertIsNone(self.manager.active)

    def test_blocking_failure_does_not_affect_background(self):
        blocking = QTask()
        bg = QTask()
        self.manager.register(blocking)
        self.manager.register(bg, blocking=False)
        blocking.abort()
        self.assertIn(bg, self.manager.background)


class TestQTaskManagerStop(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def test_stop_rewinds_queue(self):
        # After stop, tasks are reset and first is activated;
        # remaining tasks are back in the queue.
        for _ in range(3):
            self.manager.register(QTask())
        self.manager.stop()
        self.assertEqual(self.manager.queue_size, 2)

    def test_stop_resets_active_task_to_running(self):
        # stop() aborts the current task then resets and re-activates it.
        task = QTask()
        self.manager.register(task)
        self.manager.stop()
        # Task is reset to PENDING then immediately started → RUNNING,
        # but not yet stepped so active returns None.
        self.assertIsNone(self.manager.active)
        self.assertEqual(self.manager.active_raw.state, QTask.State.RUNNING)

    def test_stop_removes_background_tasks(self):
        for _ in range(2):
            self.manager.register(QTask(), blocking=False)
        self.manager.stop()
        self.assertEqual(len(self.manager.background), 0)

    def test_stop_aborts_background_tasks(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        self.manager.stop()
        self.assertEqual(task.state, QTask.State.FAILED)

    def test_stop_clears_active_reference(self):
        self.manager.register(QTask())
        self.manager.stop()
        self.assertIsNone(self.manager.active)

    def test_stop_preserves_schedule(self):
        tasks = [QTask() for _ in range(3)]
        for t in tasks:
            self.manager.register(t)
        self.manager.stop()
        self.assertEqual(len(self.manager.scheduled), 3)

    def test_stop_on_empty_manager_does_not_raise(self):
        self.manager.stop()


class TestQTaskManagerSchedule(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_scheduled_empty_initially(self):
        self.assertEqual(self.manager.scheduled, [])

    def test_registered_task_in_scheduled(self):
        task = QTask()
        self.manager.register(task)
        self.assertIn(task, self.manager.scheduled)

    def test_scheduled_preserves_order(self):
        tasks = [QTask() for _ in range(3)]
        for t in tasks:
            self.manager.register(t)
        self.assertEqual(self.manager.scheduled, tasks)

    def test_scheduled_persists_after_task_completes(self):
        task = QTask(duration=1)
        self.manager.register(task)
        self._emit(1)
        self.assertIn(task, self.manager.scheduled)

    def test_scheduled_persists_after_stop(self):
        task = QTask()
        self.manager.register(task)
        self.manager.stop()
        self.assertIn(task, self.manager.scheduled)

    def test_scheduled_property_is_copy(self):
        task = QTask()
        self.manager.register(task)
        s = self.manager.scheduled
        s.clear()
        self.assertEqual(len(self.manager.scheduled), 1)

    def test_background_tasks_not_in_scheduled(self):
        bg = QTask()
        self.manager.register(bg, blocking=False)
        self.assertNotIn(bg, self.manager.scheduled)

    def test_clear_empties_schedule(self):
        for _ in range(3):
            self.manager.register(QTask())
        self.manager.clear()
        self.assertEqual(self.manager.scheduled, [])

    def test_clear_also_stops_execution(self):
        task = QTask()
        self.manager.register(task)
        self.manager.clear()
        self.assertIsNone(self.manager.active)
        self.assertEqual(self.manager.queue_size, 0)

    def test_clear_on_empty_manager_does_not_raise(self):
        self.manager.clear()

    def test_restart_reruns_schedule(self):
        from QHOT.tasks.Delay import Delay
        self.manager.register(Delay(frames=5))
        self.manager.restart()
        self.assertEqual(len(self.manager.scheduled), 1)
        self.assertEqual(self.manager.active_raw.frames, 5)

    def test_restart_creates_fresh_instances(self):
        from QHOT.tasks.Delay import Delay
        task = Delay(frames=1)
        self.manager.register(task)
        self._emit(1)         # task completes
        self.manager.restart()
        new_task = self.manager.active_raw
        self.assertIsNot(new_task, task)
        self.assertEqual(new_task.state, QTask.State.RUNNING)

    def test_restart_on_empty_schedule_is_noop(self):
        self.manager.restart()  # should not raise
        self.assertEqual(self.manager.scheduled, [])


class TestQTaskManagerAutoReset(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_manager_paused_after_auto_reset(self):
        from QHOT.tasks.Delay import Delay
        self.manager.register(Delay(frames=1))
        self._emit(1)             # task completes → auto-reset
        self.assertTrue(self.manager.paused)

    def test_first_task_activated_after_auto_reset(self):
        from QHOT.tasks.Delay import Delay
        task = Delay(frames=1)
        self.manager.register(task)
        self._emit(1)             # task completes → auto-reset, re-activated
        self.assertIs(self.manager.active_raw, task)

    def test_task_running_after_auto_reset(self):
        from QHOT.tasks.Delay import Delay
        task = Delay(frames=1)
        self.manager.register(task)
        self._emit(1)
        self.assertEqual(task.state, QTask.State.RUNNING)

    def test_schedule_preserved_after_auto_reset(self):
        from QHOT.tasks.Delay import Delay
        task = Delay(frames=1)
        self.manager.register(task)
        self._emit(1)
        self.assertIn(task, self.manager.scheduled)

    def test_resuming_after_auto_reset_reruns_tasks(self):
        from QHOT.tasks.Delay import Delay
        from pyqtgraph.Qt import QtTest
        task = Delay(frames=1)
        self.manager.register(task)
        self._emit(1)             # first run completes → auto-reset, re-activated
        spy = QtTest.QSignalSpy(task.started)
        self.manager.pause(False)
        self._emit(1)             # second run
        self.assertGreater(len(spy), 0)

    def test_stop_leaves_manager_paused(self):
        from QHOT.tasks.Delay import Delay
        self.manager.register(Delay(frames=1))
        self.manager.stop()
        self.assertTrue(self.manager.paused)


class TestQTaskManagerBackground(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_background_task_removed_on_completion(self):
        task = QTask(duration=1)
        self.manager.register(task, blocking=False)
        self._emit(1)
        self.assertEqual(len(self.manager.background), 0)

    def test_background_task_removed_on_failure(self):
        task = QTask()
        self.manager.register(task, blocking=False)
        task.abort('test')
        self.assertEqual(len(self.manager.background), 0)

    def test_multiple_background_tasks_run_simultaneously(self):
        steps = []
        for i in range(3):
            t = QTask()
            t.process = MagicMock(
                side_effect=lambda f, _i=i: steps.append(_i))
            self.manager.register(t, blocking=False)
        self._emit(1)
        self.assertEqual(sorted(steps), [0, 1, 2])

    def test_background_runs_alongside_blocking(self):
        blocking = QTask()
        bg       = QTask()
        blocking._step = MagicMock()
        bg._step       = MagicMock()
        self.manager.register(blocking)
        self.manager.register(bg, blocking=False)
        self._emit(1)
        blocking._step.assert_called_once()
        bg._step.assert_called_once()


class TestQTaskManagerLoad(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.overlay = MagicMock()
        self.cgh     = MagicMock()
        self.dvr     = MagicMock()
        self.manager = QTaskManager(
            self.screen,
            overlay=self.overlay,
            cgh=self.cgh,
            dvr=self.dvr)

    def test_load_registers_tasks(self):
        from QHOT.tasks.ClearTraps import ClearTraps
        t1, t2 = ClearTraps(), ClearTraps()
        self.manager.register(t1)   # becomes active
        self.manager.register(t2)   # queued
        dicts = [t.to_dict() for t in self.manager.queued]
        self.manager.stop()
        self.manager.load(dicts)
        self.assertIsNotNone(self.manager.active_raw)

    def test_load_injects_overlay(self):
        from QHOT.tasks.ClearTraps import ClearTraps
        d = ClearTraps().to_dict()
        self.manager.load([d])
        self.assertIs(self.manager.active_raw.overlay, self.overlay)

    def test_load_injects_cgh(self):
        from QHOT.tasks.ClearTraps import ClearTraps
        d = ClearTraps().to_dict()
        self.manager.load([d])
        self.assertIs(self.manager.active_raw.cgh, self.cgh)

    def test_load_injects_dvr(self):
        from QHOT.tasks.ClearTraps import ClearTraps
        d = ClearTraps().to_dict()
        self.manager.load([d])
        self.assertIs(self.manager.active_raw.dvr, self.dvr)

    def test_load_preserves_task_params(self):
        from QHOT.tasks.Delay import Delay
        d = Delay(frames=99).to_dict()
        self.manager.load([d])
        self.assertEqual(self.manager.active_raw.frames, 99)

    def test_load_appends_to_existing_queue(self):
        from QHOT.tasks.ClearTraps import ClearTraps
        self.manager.register(QTask())       # becomes active
        self.manager.register(QTask())       # queued
        self.manager.load([ClearTraps().to_dict()])
        self.assertEqual(self.manager.queue_size, 2)

    def test_load_multiple_tasks_in_order(self):
        from QHOT.tasks.Delay import Delay
        dicts = [Delay(frames=i).to_dict() for i in (10, 20, 30)]
        self.manager.load(dicts)
        # first task activated (not yet stepped) — still in queued
        queued = self.manager.queued
        frames = [t.frames for t in queued]
        self.assertEqual(frames, [10, 20, 30])

    def test_load_empty_list_is_noop(self):
        self.manager.load([])
        self.assertIsNone(self.manager.active)
        self.assertEqual(self.manager.queue_size, 0)


class TestQTaskManagerRemove(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _register(self, n: int) -> list:
        tasks = [QTask() for _ in range(n)]
        for t in tasks:
            self.manager.register(t)
        return tasks

    def test_remove_pending_task_from_schedule(self):
        t1, t2 = self._register(2)
        self.manager.remove(t2)
        self.assertNotIn(t2, self.manager.scheduled)

    def test_remove_pending_task_from_queue(self):
        t1, t2 = self._register(2)
        self.manager.remove(t2)
        self.assertEqual(self.manager.queue_size, 0)

    def test_remove_active_task_is_ignored(self):
        t1, t2 = self._register(2)
        # t1 is active
        self.manager.remove(t1)
        self.assertIn(t1, self.manager.scheduled)
        self.assertIs(self.manager.active_raw, t1)

    def test_remove_active_logs_warning(self):
        t1 = self._register(1)[0]
        with self.assertLogs('QHOT.lib.tasks.QTaskManager', level='WARNING'):
            self.manager.remove(t1)

    def test_remove_task_not_in_schedule_is_noop(self):
        t1 = self._register(1)[0]
        outsider = QTask()
        self.manager.remove(outsider)   # should not raise
        self.assertEqual(len(self.manager.scheduled), 1)

    def test_remove_completed_task(self):
        from QHOT.tasks.Delay import Delay
        mgr = QTaskManager(self.screen)
        d1 = Delay(frames=0)
        d2 = Delay(frames=0)
        mgr.register(d1)
        mgr.register(d2)
        self.screen.rendered.emit()   # d1 completes
        self.assertEqual(d1.state, QTask.State.COMPLETED)
        mgr.remove(d1)
        self.assertNotIn(d1, mgr.scheduled)

    def test_remove_emits_changed(self):
        t1, t2 = self._register(2)
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.remove(t2)
        self.assertGreater(len(spy), 0)

    def test_remove_only_removes_one_task(self):
        t1, t2, t3 = self._register(3)
        self.manager.remove(t2)
        self.assertEqual(len(self.manager.scheduled), 2)
        self.assertIn(t1, self.manager.scheduled)
        self.assertIn(t3, self.manager.scheduled)


class TestQTaskManagerChanged(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _emit(self, n: int = 1) -> None:
        for _ in range(n):
            self.screen.rendered.emit()

    def test_changed_emitted_on_first_blocking_register(self):
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.register(QTask())
        self.assertGreater(len(spy), 0)

    def test_changed_emitted_on_queued_blocking_register(self):
        self.manager.register(QTask())           # activates
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.register(QTask())           # queues
        self.assertGreater(len(spy), 0)

    def test_changed_emitted_on_background_register(self):
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.register(QTask(), blocking=False)
        self.assertGreater(len(spy), 0)

    def test_changed_emitted_on_pause(self):
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.pause(True)
        self.assertEqual(len(spy), 1)

    def test_changed_not_emitted_on_repeated_pause(self):
        self.manager.pause(True)
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.pause(True)          # same state — no change
        self.assertEqual(len(spy), 0)

    def test_changed_emitted_on_resume(self):
        self.manager.pause(True)
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.pause(False)
        self.assertEqual(len(spy), 1)

    def test_changed_emitted_on_stop(self):
        self.manager.register(QTask())
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.stop()
        self.assertGreater(len(spy), 0)

    def test_changed_emitted_when_task_completes(self):
        self.manager.register(QTask(duration=1))
        spy = QtTest.QSignalSpy(self.manager.changed)
        self._emit(1)
        self.assertGreater(len(spy), 0)

    def test_changed_emitted_when_background_completes(self):
        self.manager.register(QTask(duration=1), blocking=False)
        spy = QtTest.QSignalSpy(self.manager.changed)
        self._emit(1)
        self.assertGreater(len(spy), 0)

    def test_changed_emitted_when_task_fails(self):
        task = QTask()
        self.manager.register(task)
        spy = QtTest.QSignalSpy(self.manager.changed)
        task.abort('test')
        self.assertGreater(len(spy), 0)


class TestQTaskManagerReorder(unittest.TestCase):

    def setUp(self):
        self.screen  = MockScreen()
        self.manager = QTaskManager(self.screen)

    def _register(self, n: int) -> list:
        tasks = [QTask() for _ in range(n)]
        for t in tasks:
            self.manager.register(t)
        return tasks

    def test_reorder_changes_schedule_order(self):
        t1, t2, t3 = self._register(3)
        self.manager.reorder([t3, t1, t2])
        self.assertEqual(self.manager.scheduled, [t3, t1, t2])

    def test_reorder_changes_queue_order(self):
        t1, t2, t3 = self._register(3)
        # t1 is active (running), t2 and t3 are pending in queue
        self.manager.reorder([t1, t3, t2])
        self.assertEqual(list(self.manager._queue), [t3, t2])

    def test_reorder_active_task_excluded_from_queue(self):
        t1, t2, t3 = self._register(3)
        # t1 is currently active (_current); reorder puts t1 last
        self.manager.reorder([t2, t3, t1])
        # t1 is still _current, not re-enqueued
        self.assertIs(self.manager.active_raw, t1)
        self.assertNotIn(t1, list(self.manager._queue))

    def test_reorder_with_wrong_tasks_ignored(self):
        t1, t2 = self._register(2)
        outsider = QTask()
        original = self.manager.scheduled[:]
        self.manager.reorder([t1, outsider])
        self.assertEqual(self.manager.scheduled, original)

    def test_reorder_emits_changed(self):
        t1, t2 = self._register(2)
        spy = QtTest.QSignalSpy(self.manager.changed)
        self.manager.reorder([t2, t1])
        self.assertGreater(len(spy), 0)

    def test_reorder_completed_tasks_stay_in_place(self):
        screen = self.screen
        t1, t2, t3 = self._register(3)
        # Complete t1 by stepping it (t1 has duration=None, won't complete)
        # Use a zero-duration task instead
        from QHOT.tasks.Delay import Delay
        mgr = QTaskManager(screen)
        d1 = Delay(frames=0)
        d2 = Delay(frames=0)
        d3 = Delay(frames=0)
        mgr.register(d1)
        mgr.register(d2)
        mgr.register(d3)
        screen.rendered.emit()   # d1 completes, d2 activates
        self.assertEqual(d1.state, QTask.State.COMPLETED)
        # Reorder d2 and d3 (both PENDING/RUNNING)
        mgr.reorder([d1, d3, d2])
        self.assertEqual(mgr.scheduled, [d1, d3, d2])

    def test_reorder_preserves_same_set(self):
        t1, t2, t3 = self._register(3)
        self.manager.reorder([t3, t2, t1])
        self.assertEqual(set(id(t) for t in self.manager.scheduled),
                         {id(t1), id(t2), id(t3)})


if __name__ == '__main__':
    unittest.main()



'''Unit tests for LoadTraps.'''
import json
import tempfile
import unittest
from unittest.mock import MagicMock

from pyqtgraph.Qt import QtWidgets

from QHOT.lib.tasks.QTask import QTask
from QHOT.tasks.LoadTraps import LoadTraps

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


def _run(task):
    task._start()
    task._step()


class TestLoadTrapsInit(unittest.TestCase):

    def test_registered_in_registry(self):
        self.assertIn('LoadTraps', QTask._registry)

    def test_duration_is_zero(self):
        self.assertEqual(LoadTraps().duration, 0)

    def test_default_filename_empty(self):
        self.assertEqual(LoadTraps().filename, '')

    def test_explicit_filename_stored(self):
        task = LoadTraps(filename='/tmp/traps.json')
        self.assertEqual(task.filename, '/tmp/traps.json')

    def test_parameters_declared(self):
        names = [p['name'] for p in LoadTraps.parameters]
        self.assertIn('filename', names)


class TestLoadTrapsExecution(unittest.TestCase):

    def setUp(self):
        self.overlay = MagicMock()
        self.trap_data = [{'type': 'QTweezer', 'x': 10., 'y': 20., 'z': 0.}]
        self.tmp = tempfile.NamedTemporaryFile(
            suffix='.json', mode='w', delete=False)
        json.dump(self.trap_data, self.tmp)
        self.tmp.close()

    def test_from_list_called_with_file_contents(self):
        task = LoadTraps(overlay=self.overlay, filename=self.tmp.name)
        _run(task)
        self.overlay.from_list.assert_called_once_with(self.trap_data)

    def test_empty_filename_does_not_call_from_list(self):
        task = LoadTraps(overlay=self.overlay, filename='')
        _run(task)
        self.overlay.from_list.assert_not_called()

    def test_empty_filename_still_completes(self):
        task = LoadTraps(overlay=self.overlay, filename='')
        _run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)

    def test_completes_after_one_step(self):
        task = LoadTraps(overlay=self.overlay, filename=self.tmp.name)
        _run(task)
        self.assertEqual(task.state, QTask.State.COMPLETED)


class TestLoadTrapsSerialization(unittest.TestCase):

    def test_to_dict_includes_type(self):
        self.assertEqual(LoadTraps().to_dict()['type'], 'LoadTraps')

    def test_round_trip(self):
        task = LoadTraps(filename='/tmp/traps.json')
        restored = QTask.from_dict(task.to_dict())
        self.assertIsInstance(restored, LoadTraps)
        self.assertEqual(restored.filename, '/tmp/traps.json')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



'''Unit tests for QTaskTree.'''
import unittest

from pyqtgraph.Qt import QtWidgets

from QHOT.lib.tasks.QTask import QTask
from QHOT.lib.tasks.QTaskTree import QTaskTree

app = QtWidgets.QApplication.instance() or QtWidgets.QApplication([])


class _NoParams(QTask):
    '''Task with no declared parameters.'''
    parameters = []


class _TwoParams(QTask):
    '''Task with two declared parameters.'''
    parameters = [
        dict(name='alpha', type='float', value=1.0),
        dict(name='beta',  type='int',   value=10),
    ]

    def __init__(self, alpha: float = 1.0, beta: int = 10, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.beta  = beta


class TestQTaskTreeInit(unittest.TestCase):

    def test_task_property_returns_task(self):
        task = _TwoParams()
        tree = QTaskTree(task)
        self.assertIs(tree.task, task)

    def test_params_group_named_after_task_class(self):
        task = _TwoParams()
        tree = QTaskTree(task)
        root = tree.invisibleRootItem()
        self.assertEqual(root.child(0).text(0), '_TwoParams')

    def test_param_count_matches_parameters_spec(self):
        task = _TwoParams()
        tree = QTaskTree(task)
        root = tree.invisibleRootItem()
        group = root.child(0)
        self.assertEqual(group.childCount(), len(_TwoParams.parameters))

    def test_empty_parameters_produces_empty_group(self):
        task = _NoParams()
        tree = QTaskTree(task)
        root = tree.invisibleRootItem()
        group = root.child(0)
        self.assertEqual(group.childCount(), 0)

    def test_initial_values_from_instance_attrs(self):
        task = _TwoParams(alpha=3.5, beta=42)
        tree = QTaskTree(task)
        self.assertAlmostEqual(tree._params.child('alpha').value(), 3.5)
        self.assertEqual(tree._params.child('beta').value(), 42)

    def test_initial_values_not_just_class_defaults(self):
        task = _TwoParams(alpha=99.0)
        tree = QTaskTree(task)
        self.assertAlmostEqual(tree._params.child('alpha').value(), 99.0)


class TestQTaskTreeSync(unittest.TestCase):

    def test_editing_param_updates_task_attribute(self):
        task = _TwoParams(alpha=1.0)
        tree = QTaskTree(task)
        tree._params.child('alpha').setValue(7.5)
        self.assertAlmostEqual(task.alpha, 7.5)

    def test_editing_int_param_updates_task_attribute(self):
        task = _TwoParams(beta=10)
        tree = QTaskTree(task)
        tree._params.child('beta').setValue(99)
        self.assertEqual(task.beta, 99)

    def test_multiple_edits_all_propagate(self):
        task = _TwoParams(alpha=0.0, beta=0)
        tree = QTaskTree(task)
        tree._params.child('alpha').setValue(2.5)
        tree._params.child('beta').setValue(5)
        self.assertAlmostEqual(task.alpha, 2.5)
        self.assertEqual(task.beta, 5)

    def test_ignore_sync_prevents_update(self):
        task = _TwoParams(alpha=1.0)
        tree = QTaskTree(task)
        tree._ignoreSync = True
        tree._params.child('alpha').setValue(99.0)
        self.assertAlmostEqual(task.alpha, 1.0)

    def test_two_trees_on_same_task_both_sync(self):
        task = _TwoParams(alpha=1.0)
        tree1 = QTaskTree(task)
        tree2 = QTaskTree(task)
        tree1._params.child('alpha').setValue(2.0)
        self.assertAlmostEqual(task.alpha, 2.0)
        tree2._params.child('alpha').setValue(3.0)
        self.assertAlmostEqual(task.alpha, 3.0)


if __name__ == '__main__':
    unittest.main()
